import logging

import openpaperwork_core


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return ['sync_on_start']

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'doc_labels',
                'defaults': ['paperwork_backend.model.labels'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def init(self, core):
        super().init(core)
        setting = self.core.call_success(
            "config_build_simple", "GUI", "sync_on_start",
            default_value_func=lambda: True
        )
        self.core.call_all("config_register", "sync_on_start", setting)

    def on_gtk_initialized(self):
        r = self.core.call_success("config_get", "sync_on_start")
        if r:
            LOGGER.info("Starting synchronization ...")
            self.core.call_all("transaction_sync_all")
        else:
            LOGGER.info(
                "Synchronization on start is disabled --> Just loading labels"
            )
            promises = []
            self.core.call_all("label_load_all", promises)
            promise = promises[0]
            for p in promises[1:]:
                promise = promise.then(p)
            # use transaction_schedule to make sure that document imports
            # are not done at the same time.
            self.core.call_one("transaction_schedule", promise)



import logging
import re

import openpaperwork_core

from .. import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    RE_SPLIT_SOURCE_NAME = [
        re.compile(r'([()])'),
        re.compile(r'(\W)'),
    ]

    def __init__(self):
        super().__init__()
        # Need the l10n plugin to be loaded first before getting the
        # translations
        self.keywords = {}

    def get_interfaces(self):
        return ['i18n_scanner']

    def get_deps(self):
        return [
            {
                'interface': 'l10n',
                'defaults': ['openpaperwork_core.l10n.python'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.keywords = {
            "centrally aligned": _("centrally aligned"),
            "feeder": _("Feeder"),
            "flatbed": _("Flatbed"),
            "left aligned": _("left aligned"),
            "right aligned": _("right aligned"),
        }

    def i18n_scanner_source(self, source_name):
        original = source_name

        for regex in self.RE_SPLIT_SOURCE_NAME:
            source_name = regex.split(source_name)
            for i in range(0, len(source_name)):
                if source_name[i].lower() in self.keywords:
                    source_name[i] = self.keywords[source_name[i].lower()]
            source_name = "".join(source_name)

        LOGGER.debug("I18n: %s --> %s", original, source_name)
        return source_name



import logging

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -500

    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return [
            'gtk_settings_ocr_langs',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'i18n_lang',
                'defaults': ['paperwork_backend.i18n.pycountry'],
            },
            {
                'interface': 'ocr_settings',
                'defaults': ['paperwork_backend.pyocr'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
        ]

    def complete_ocr_settings(self, parent_widget_tree):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.settings.ocr", "selector_popover.glade"
        )

        active_langs = set(self.core.call_success("ocr_get_active_langs"))
        LOGGER.info("Looking for available OCR languages ...")
        all_langs = self.core.call_success("ocr_get_available_langs")
        LOGGER.info("Found %d languages. Translating ...", len(all_langs))
        all_langs = [
            (
                lang,
                self.core.call_success(
                    "i18n_lang_iso639_3_to_full", lang
                )
            )
            for lang in all_langs
            # Skip Tesseract data file for page orientation guessing
            if lang != 'osd'
        ]
        all_langs.sort(key=lambda lang: lang[1])

        LOGGER.info("OCR languages: %s", all_langs)
        LOGGER.info("Active OCR languages: %s", active_langs)

        box_parent = widget_tree.get_object("ocr_selector_box")

        for lang in all_langs:
            w_tree = self.core.call_success(
                "gtk_load_widget_tree",
                "paperwork_gtk.settings.ocr", "selector_popover_box.glade"
            )
            check = w_tree.get_object("ocr_selector_box")
            check.set_label(lang[1])
            check.set_active(lang[0] in active_langs)
            check.connect('toggled', self._on_toggle, lang[0])
            box_parent.pack_start(check, expand=False, fill=True, padding=0)
        LOGGER.info("OCR selector ready")

        popover = widget_tree.get_object("ocr_selector")
        parent_widget_tree.get_object("ocr_langs").set_popover(popover)

    def _on_toggle(self, checkbox, lang):
        active_langs = self.core.call_success("ocr_get_active_langs")
        lang_enabled = lang in active_langs
        LOGGER.info("Language toggled: {} ({} -> {})".format(
            lang, lang_enabled, not lang_enabled
        ))
        if lang_enabled:
            active_langs.remove(lang)
        else:
            active_langs.append(lang)
        self.core.call_all("ocr_set_active_langs", active_langs)



import logging

try:
    import gi
    gi.require_version('Pango', '1.0')
    gi.require_version('PangoCairo', '1.0')
    from gi.repository import Pango
    from gi.repository import PangoCairo
    PANGO_AVAILABLE = True
except (ImportError, ValueError):
    PANGO_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_core.promise

from ..... import _


LOGGER = logging.getLogger(__name__)
DELAY = 0.1


class NBox(object):
    """
    Chained boxes. Useful for some plugins like boxes.selection.
    """
    def __init__(self, box, index):
        self.box = box
        self.next = None
        self.index = index

    def __lt__(self, other):
        if self.index < other.index:
            return True
        return (self.box.position < other.box.position)


class Plugin(openpaperwork_core.PluginBase):
    """
    Load the boxes on the pages and notify them to other plugins.
    Do nothing else. See other plugins in
    paperwork_gtk.mainwindow.docview.pageview.boxes to have something actually
    happening.
    """

    def __init__(self):
        super().__init__()
        self.cache = {}
        self.running_promises = {}
        self.nb_to_load = 0
        self.nb_loaded = 0

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_pageview_boxes',
        ]

    def get_deps(self):
        return [
            {
                "interface": "gtk_pageview",
                "defaults": ["paperwork_gtk.mainwindow.docview.pageview"],
            },
            {
                'interface': 'spatial_index',
                'defaults': ['openpaperwork_core.spatial.rtree'],
            },
            {
                'interface': 'work_queue',
                'defaults': ['openpaperwork_core.work_queue.default'],
            },
        ]

    def chkdeps(self, out: dict):
        if not PANGO_AVAILABLE:
            out['pango'].update(openpaperwork_core.deps.PANGO)

    def _upd_progress(self):
        if self.nb_to_load <= self.nb_loaded or self.nb_to_load == 0:
            self.core.call_all("on_progress", "boxes", 1.0)
            self.nb_to_load = 0
            self.nb_loaded = 0
            return
        self.core.call_all(
            "on_progress", "boxes",
            self.nb_loaded / self.nb_to_load, _("Loading text ...")
        )

    def doc_close(self):
        self.cache = {}
        self.nb_to_load = 0
        self.nb_loaded = 0
        self._upd_progress()

    def doc_open(self, *args, **kwargs):
        self.doc_close()

    def _index_boxes(self, boxes):
        # Tesseract seems (seemed ?) to have a bug: boxes taking the whole
        # page. --> we remove them.
        # Also we strip empty boxes.
        boxes = [
            line_box for line_box in boxes
            if line_box.position[0][0] > 0 or line_box.position[0][1] > 0
        ]
        for line_box in boxes:
            line_box.word_boxes = [
                word_box for word_box in line_box.word_boxes
                if ((word_box.position[0][0] > 0 or
                    word_box.position[0][1] > 0) and
                    word_box.content.strip() != "")
            ]

        # chain the boxes
        chained_boxes = []
        pbox = None
        index = 0
        for line in boxes:
            for word in line.word_boxes:
                new_box = NBox(word, index)
                index += 1
                if pbox is not None:
                    pbox.next = new_box
                    chained_boxes.append(pbox)
                pbox = new_box
        if pbox is not None:
            chained_boxes.append(pbox)

        # and then index them
        spatial_index = self.core.call_success(
            "spatial_indexer_get", [
                (b.box.position, b)
                for b in chained_boxes
            ]
        )

        return (boxes, spatial_index)

    def _set_boxes(self, boxes, page):
        (boxes, spatial_index) = boxes
        ref = (page.doc_id, page.page_idx)
        self.cache[ref] = (boxes, spatial_index)
        return (boxes, spatial_index)

    def pageview_get_boxes_by_id(self, doc_id, page_idx):
        ref = (doc_id, page_idx)
        if ref not in self.cache:
            return None
        return self.cache[ref][0]

    def pageview_get_indexed_boxes_by_id(self, doc_id, page_idx):
        ref = (doc_id, page_idx)
        if ref not in self.cache:
            return None
        return self.cache[ref][1]

    def on_page_visibility_changed(self, page, visible):
        ref = (page.doc_id, page.page_idx)
        if not visible:
            promise = self.running_promises.pop(ref, None)
            if promise is not None:
                self.core.call_all("work_queue_cancel", "page_loader", promise)
                self.nb_to_load -= 1
                self._upd_progress()
            if ref in self.cache:
                self.cache.pop(ref)
            return

        if ref in self.cache:
            return

        # even they are not yet loaded, they will soon be --> we can mark them
        # as loaded
        self.cache[ref] = (None, None)

        # Gives back a bit of CPU time to GTK so the GUI remains
        # usable
        promise = openpaperwork_core.promise.DelayPromise(self.core, DELAY)

        promise = promise.then(
            LOGGER.debug,
            "Loading boxes of %s p%d", page.doc_id, page.page_idx
        )
        # drop the returned value
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(openpaperwork_core.promise.ThreadedPromise(
            self.core, self.core.call_success,
            args=("page_get_boxes_by_url", page.doc_url, page.page_idx,)
        ))
        promise = promise.then(openpaperwork_core.promise.ThreadedPromise(
            self.core, lambda boxes=[]: self._index_boxes(boxes)
        ))
        promise = promise.then(lambda boxes: self._set_boxes(boxes, page))
        promise = promise.then(lambda boxes: self.core.call_all(
            # boxes => (boxes, spatial_index)
            "on_page_boxes_loaded", page, boxes[0], boxes[1]
        ))

        def stop_promise_tracking(*args, **kwargs):
            if ref in self.running_promises:
                self.nb_loaded += 1
                self._upd_progress()
                self.running_promises.pop(ref)

        promise = promise.then(stop_promise_tracking)

        self.nb_to_load += 1
        self._upd_progress()
        self.running_promises[ref] = promise

        # piggy back page loader work queue, but with a low priority
        self.core.call_success(
            "work_queue_add_promise", "page_loader", promise, priority=-10
        )

    def on_page_boxes_loaded(self, page, boxes, spatial_index):
        LOGGER.info(
            "Page %s %d: %d line boxes loaded",
            page.doc_id, page.page_idx, len(boxes)
        )

    def _paint_txt(self, cairo_ctx, txt, x, y, w, h):
        cairo_ctx.set_source_rgb(1.0, 1.0, 1.0)
        cairo_ctx.rectangle(x, y, w, h)
        cairo_ctx.fill()

        layout = PangoCairo.create_layout(cairo_ctx)
        layout.set_text(txt, -1)
        txt_size = layout.get_size()
        if 0 in txt_size:
            return

        cairo_ctx.save()
        try:
            txt_factor = min(
                float(w) * Pango.SCALE / txt_size[0],
                float(h) * Pango.SCALE / txt_size[1],
            )
            cairo_ctx.set_source_rgb(0, 0, 0)
            cairo_ctx.translate(x, y)

            # make the text use the whole box space
            cairo_ctx.scale(txt_factor, txt_factor)

            PangoCairo.update_layout(cairo_ctx, layout)
            PangoCairo.show_layout(cairo_ctx, layout)
        finally:
            cairo_ctx.restore()

    def _page_draw_box(
            self, cairo_ctx, page, box_position,
            border_color, border_width=2,
            box_content=None):
        zoom = page.zoom
        ((tl_x, tl_y), (br_x, br_y)) = box_position
        tl_x *= zoom
        tl_y *= zoom
        br_x *= zoom
        br_y *= zoom
        w = br_x - tl_x
        h = br_y - tl_y

        if box_content is not None:
            self._paint_txt(cairo_ctx, box_content, tl_x, tl_y, w, h)

        cairo_ctx.save()
        try:

            cairo_ctx.set_source_rgb(
                border_color[0], border_color[1], border_color[2]
            )
            cairo_ctx.set_line_width(border_width)
            cairo_ctx.rectangle(
                tl_x - (border_width / 2),
                tl_y - (border_width / 2),
                w + border_width,
                h + border_width
            )
            cairo_ctx.stroke()
        finally:
            cairo_ctx.restore()

    def page_draw_box(self, *args, **kwargs):
        # WORKAROUND(Jflesch): with some malformed PDF file, we get unexpected
        # exceptions. See:
        # https://gitlab.gnome.org/World/OpenPaperwork/paperwork/-/issues/913
        # https://gitlab.freedesktop.org/poppler/poppler/-/issues/1020
        try:
            self._page_draw_box(*args, **kwargs)
        except Exception as exc:
            LOGGER.error("page_draw_box() failed", exc_info=exc)



import datetime
import locale
import logging

import openpaperwork_core
import openpaperwork_core.promise
import paperwork_backend.model.pdf

from ... import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10000

    def __init__(self):
        super().__init__()
        self.doc_urls_to_names_to_urls = {}
        self.doc_urls_to_names = {}

        # in case document is edited by user
        self.new_docs = []
        self.doc_urls_to_new_docs = {}

        self.thumbnails = {}
        # At this point, translations are not yet available
        self.help_files = ()
        self.label = ()

    def get_interfaces(self):
        return [
            'doc_labels',
            'document_storage',
            'help_documents',
            'thumbnail',
        ]

    def get_deps(self):
        return [
            # Dependency not declared to avoid dependency loop
            # {
            #     'interface': 'new_doc',
            #     'defaults': ['paperwork_gtk.new_doc'],
            # },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'page_img',
                'defaults': ['paperwork_backend.model.pdf'],
            },
            {
                'interface': 'pillow',
                'defaults': [
                    'openpaperwork_core.pillow.img',
                    'paperwork_backend.pillow.pdf',
                ],
            },
            {
                'interface': 'resources',
                'defaults': ['openpaperwork_core.resources.setuptools'],
            },
            {
                'interface': 'thumbnailer',
                'defaults': ['paperwork_backend.model.thumbnail'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.help_files = (
            (_("Introduction"), "intro"),
            (_("User manual"), "usage"),
        )
        self.label = (_("Documentation"), "#ffffffffffff")

    def help_get_files(self):
        return self.help_files

    def help_get_file(self, name):
        lang = "en"
        try:
            lang = locale.getdefaultlocale()[0]
            if lang is None:
                lang = "en"
                LOGGER.warning("Failed to get default locale !")
            else:
                lang = lang[:2]
                LOGGER.info("User language: %s", lang)
        except Exception as exc:
            LOGGER.error(
                "get_documentation(): Failed to figure out locale."
                " Will default to English",
                exc_info=exc
            )
        if lang == "en":
            docs = [name + '.pdf']
        else:
            docs = ['translated_{}_{}.pdf'.format(name, lang), name + ".pdf"]

        for doc in docs:
            path_ctx = self.core.call_success(
                "resources_get_file", "paperwork_gtk.model.help.out", doc
            )
            if path_ctx is None:
                LOGGER.warning("No documentation '%s' found", doc)
            else:
                with path_ctx as path:
                    LOGGER.info("Documentation '%s': %s", doc, path)
                    url = self.core.call_success("fs_safe", path)
                    self.doc_urls_to_names_to_urls[name] = url
                    self.doc_urls_to_names[url] = name
                return url
        LOGGER.error("Failed to find documentation '%s'", name)
        return None

    def doc_id_to_url(self, doc_id, existing=True):
        if not doc_id.startswith("help_"):
            return None
        name = doc_id[len("help_"):]
        return self.help_get_file(name)

    def doc_get_date_by_id(self, doc_id):
        if not doc_id.startswith("help_"):
            return None
        return datetime.datetime.now()

    def thumbnail_get_doc_promise(self, doc_url):
        if doc_url not in self.doc_urls_to_names:
            return None
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self.thumbnail_get_doc, args=(doc_url,)
        )

    def thumbnail_get_doc(self, doc_url):
        return self.thumbnail_get_page(doc_url, page_idx=0)

    def thumbnail_get_page(self, doc_url, page_idx):
        if doc_url not in self.doc_urls_to_names:
            return None
        if page_idx != 0:
            return None

        if doc_url in self.thumbnails:
            url = self.thumbnais[doc_url]
            return self.core.call_success("url_to_pillow", url)

        page_url = self.core.call_success(
            "page_get_img_url", doc_url, page_idx
        )
        assert page_url is not None
        img = self.core.call_success("url_to_pillow", page_url)
        img = self.core.call_success("thumbnail_from_img", img)

        (self.thumbnail_url, fd) = self.core.call_success(
            "fs_mktemp", prefix="thumbnail_help_intro", suffix=".jpeg",
            mode="wb"
        )
        fd.close()
        self.core.call_success(
            "pillow_to_url", img, self.thumbnail_url,
            format='JPEG', quality=0.85
        )
        return img

    def _get_doc_url(self, doc_url, write=False):
        # HACK(Jflesch):
        # This document is read-only, but Paperwork doesn't really
        # support read-only documents.
        # --> discretely copy the document and switch to the new document.
        new_doc_url = self.doc_urls_to_new_docs.get(doc_url, None)
        if new_doc_url is not None:
            return new_doc_url
        if not write:
            return doc_url
        (new_doc_id, new_doc_url) = self.core.call_success("get_new_doc")
        assert new_doc_url not in self.doc_urls_to_names
        self.doc_urls_to_new_docs[doc_url] = new_doc_url
        self.new_docs.append(new_doc_id)
        self.core.call_success("fs_mkdir_p", new_doc_url)
        self.core.call_success(
            "fs_copy",
            doc_url,  # reminder: doc_url points directly to a PDF here
            self.core.call_success(
                "fs_join",
                new_doc_url, paperwork_backend.model.pdf.PDF_FILENAME
            )
        )
        self.core.call_success(
            "doc_add_label_by_url", new_doc_url, *self.label
        )
        self.core.call_success(
            "mainloop_execute", self.core.call_all,
            "doc_open", new_doc_id, new_doc_url
        )
        return new_doc_url

    def page_get_img_url(self, doc_url, page_idx, write=False, **kwargs):
        if doc_url not in self.doc_urls_to_names:
            return None
        doc_url = self._get_doc_url(doc_url, write)
        if doc_url in self.doc_urls_to_names:
            return None
        return self.core.call_success(
            "page_get_img_url", doc_url, page_idx,
            write=write, **kwargs
        )

    def page_set_boxes_by_url(self, doc_url, page_idx, *args, **kwargs):
        if doc_url not in self.doc_urls_to_names:
            return None
        doc_url = self._get_doc_url(doc_url, write=True)
        return self.core.call_success(
            "page_set_boxes_by_url", doc_url, page_idx,
            *args, **kwargs
        )

    def transaction_simple_promise(self, changes: list):
        for new_doc_id in self.new_docs:
            changes.append(('add', new_doc_id))
        self.new_docs = []

    def doc_get_mtime_by_url(self, doc_url):
        if doc_url not in self.doc_urls_to_names:
            return
        return datetime.datetime(year=1971, month=1, day=1).timestamp()

    def doc_has_labels_by_url(self, doc_url):
        if doc_url not in self.doc_urls_to_names:
            return None
        return True

    def doc_get_labels_by_url(self, out: set, doc_url):
        if doc_url not in self.doc_urls_to_names:
            return
        out.add(self.label)

    def doc_get_labels_by_url_promise(self, out: list, doc_url):
        if doc_url not in self.doc_urls_to_names:
            return

        def get_labels(labels=None):
            if labels is None:
                labels = set()
            labels.add(self.label)
            return labels

        promise = openpaperwork_core.promise.Promise(
            self.core, get_labels
        )
        out.append(promise)

    def doc_remove_label_by_url(self, doc_url, *args, **kwargs):
        if doc_url not in self.doc_urls_to_names:
            return None
        doc_url = self._get_doc_url(doc_url, write=True)
        return self.core.call_success(
            "doc_remove_label_by_url", doc_url, *args, **kwargs
        )

    def doc_add_label_by_url(self, doc_url, *args, **kwargs):
        if doc_url not in self.doc_urls_to_names:
            return None
        doc_url = self._get_doc_url(doc_url, write=True)
        return self.core.call_success(
            "doc_add_label_by_url", doc_url, *args, **kwargs
        )

    def doc_set_extra_text_by_url(self, doc_url, *args, **kwargs):
        if doc_url not in self.doc_urls_to_names:
            return None
        doc_url = self._get_doc_url(doc_url, write=True)
        return self.core.call_success(
            "doc_set_extra_text_by_url", doc_url, *args, **kwargs
        )

    def help_labels_get_all(self, out: set):
        out.add(self.label)



import logging

import openpaperwork_core
import openpaperwork_core.promise

LOGGER = logging.getLogger(__name__)


class DocPropertiesUpdate(object):
    def __init__(self, doc_id, multiple_docs=False):
        self.doc_id = doc_id
        self.multiple_docs = multiple_docs

        self.new_docs = set()
        self.upd_docs = set()
        self.del_docs = set()


class DocPropertiesEditor(object):
    def __init__(self, name, plugin, multiple_docs=False):
        self.name = name
        self.core = plugin.core
        self.plugin = plugin
        self.multiple_docs = multiple_docs

        self.active_docs = set()

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docproperties", "docproperties.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        self.widget_tree.get_object("docproperties_back").connect(
            "clicked", self._apply
        )
        self.widget_tree.get_object("docproperties_cancel").connect(
            "clicked", self._cancel
        )

        self.core.call_all(
            "mainwindow_add", side="left", name="docproperties_" + self.name,
            prio=0,
            header=self.widget_tree.get_object("docproperties_header"),
            body=self.widget_tree.get_object("docproperties_body"),
        )

        self.core.call_one(
            "mainloop_schedule", self._build_doc_properties,
            multiple_docs
        )

    def show(self):
        self.core.call_all(
            "mainwindow_show", side="left", name="docproperties_" + self.name
        )

    def hide(self):
        self.core.call_all("mainwindow_back", side="left")

    def _build_doc_properties(self, multiple_docs):
        components = []
        self.core.call_all(
            "doc_properties_components_get", components,
            multiple_docs=multiple_docs
        )
        for component in components:
            self.widget_tree.get_object("docproperties_box").pack_start(
                component, expand=False, fill=True, padding=0
            )

    def doc_open(self, doc_id, doc_url):
        self.active_docs = {(doc_id, doc_url)}
        self.core.call_all(
            "doc_properties_components_set_active_doc", doc_id, doc_url
        )

    def docs_open(self, docs):
        self.active_docs = docs
        self.core.call_all(
            "doc_properties_components_set_active_docs", docs
        )

    def _open_doc(self, upd):
        active_docs = {doc[0] for doc in self.active_docs}
        if upd.doc_id is not None and upd.doc_id not in active_docs:
            doc_url = self.core.call_success("doc_id_to_url", upd.doc_id)
            if doc_url is None:
                return
            self.core.call_all("doc_open", upd.doc_id, doc_url)

    def _apply(self, *args, **kwargs):
        LOGGER.info(
            "Changes validated by the user (multiple_docs=%s)",
            self.multiple_docs
        )
        upd = DocPropertiesUpdate(
            list(self.active_docs)[0][0], self.multiple_docs
        )

        promise = openpaperwork_core.promise.ThreadedPromise(
            self.core, self.core.call_all,
            args=("doc_properties_components_apply_changes", upd)
        )
        # drop call_all return value
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(self._open_doc, upd)
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(
            self.core.call_all, "search_update_document_list"
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(self._upd_index, upd)
        promise.schedule()

        self.core.call_all("close_doc_properties")

    def _cancel(self, *args, **kwargs):
        LOGGER.info("Changes cancelled by the user")
        self.core.call_all("doc_properties_components_cancel_changes")
        self.core.call_all("close_doc_properties")

    def _upd_index(self, upd):
        total = len(upd.new_docs) + len(upd.upd_docs) + len(upd.del_docs)

        if total <= 0:
            LOGGER.info("Document %s not modified. Nothing to do", upd.doc_id)
            return

        LOGGER.info(
            "Document %s modified. %d documents impacted", upd.doc_id, total
        )

        changes = []
        for doc_id in upd.new_docs:
            changes.append(('add', doc_id))
        for doc_id in upd.upd_docs:
            changes.append(('upd', doc_id))
        for doc_id in upd.del_docs:
            changes.append(('del', doc_id))
        self.core.call_success("transaction_simple", changes)

    def docproperties_scroll_to_last(self):
        scroll = self.widget_tree.get_object("docproperties_body")
        vadj = scroll.get_vadjustment()
        vadj.set_value(vadj.get_upper())


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.editors = {}

    def get_interfaces(self):
        return ['gtk_doc_properties']

    def get_deps(self):
        return [
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_mainwindow',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.editors = {
            'single_doc': DocPropertiesEditor(
                "single", self, multiple_docs=False
            ),
            'multiple_docs': DocPropertiesEditor(
                "multiple", self, multiple_docs=True
            ),
        }

        self.active_editor = None

    def doc_open(self, doc_id, doc_url):
        if self.active_editor is None:
            return
        e = self.active_editor
        e.doc_open(doc_id, doc_url)

    def open_doc_properties(self, doc_id, doc_url):
        assert self.active_editor is None
        e = self.editors['single_doc']
        self.active_editor = e
        e.doc_open(doc_id, doc_url)
        e.show()

    def open_docs_properties(self, docs):
        assert self.active_editor is None
        e = self.editors['multiple_docs']
        self.active_editor = e
        e.docs_open(docs)
        e.show()

    def close_doc_properties(self):
        assert self.active_editor is not None
        self.active_editor.hide()
        self.active_editor = None

    def docproperties_scroll_to_last(self):
        for e in self.editors.values():
            e.docproperties_scroll_to_last()



#!/usr/bin/python3

import logging

import openpaperwork_core

EXTRA_TEXT_FILENAME = 'extra.txt'

LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            "doc_text",
            "extra_text",
        ]

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
        ]

    def doc_internal_get_mtime_by_url(self, out: list, doc_url):
        extra_url = self.core.call_success(
            "fs_join", doc_url, EXTRA_TEXT_FILENAME
        )
        if self.core.call_success("fs_exists", extra_url) is None:
            return
        out.append(self.core.call_success("fs_get_mtime", extra_url))

    def doc_get_text_by_url(self, out: list, doc_url):
        self.doc_get_extra_text_by_url(out, doc_url)

    def doc_get_extra_text_by_url(self, out: list, doc_url):
        extra_url = self.core.call_success(
            "fs_join", doc_url, EXTRA_TEXT_FILENAME
        )
        if self.core.call_success("fs_exists", extra_url) is None:
            return
        with self.core.call_success("fs_open", extra_url, 'r') as fd:
            out.append(fd.read())

    def doc_set_extra_text_by_url(self, doc_url, text):
        if not self.core.call_success("fs_isdir", doc_url):
            # Can happen on integrated documentation PDF files
            LOGGER.warning(
                "%s is not a directory. Cannot set extra text",
                doc_url
            )
            return
        extra_url = self.core.call_success(
            "fs_join", doc_url, EXTRA_TEXT_FILENAME
        )
        with self.core.call_success("fs_open", extra_url, 'w') as fd:
            fd.write(text)



"""
To use it:

```sh
paperwork-cli plugins add \
    paperwork_backend.guesswork.label.sklearn.compute_backlog
paperwork-cli compute_sklearn_label_guessing_backlog
```
"""

import functools
import itertools
import time

import sklearn.naive_bayes
import sklearn.feature_extraction.text

import openpaperwork_core


def pairwise(iterable):
    (a, b) = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


class Baye(object):
    def __init__(self):
        self.data = []
        self.targets = []

        self.sklearn_count_vectorizer = None
        self.sklearn_tfid_transformer = None
        self.sklearn_classifier = None

        self.reset()

    def reset(self):
        self.sklearn_count_vectorizer = \
            sklearn.feature_extraction.text.CountVectorizer()
        self.sklearn_tfid_transformer = \
            sklearn.feature_extraction.text.TfidfTransformer(use_idf=False)
        self.sklearn_classifier = sklearn.naive_bayes.GaussianNB()

    def fit(self, category, text):
        text = text.strip()
        if len(text) <= 0:
            return

        category = (1 if category == 'yes' else 0)
        self.data.append(text)
        self.targets.append(category)

    def seal(self):
        counts = self.sklearn_count_vectorizer.fit_transform(self.data)
        tfid = self.sklearn_tfid_transformer.fit_transform(counts)
        self.sklearn_classifier.fit(tfid.toarray(), self.targets)

    def score(self, text, label):
        counts = self.sklearn_count_vectorizer.transform([text])
        tfid = self.sklearn_tfid_transformer.transform(counts)
        predicted = self.sklearn_classifier.predict_proba(tfid.toarray())
        r = {
            "no": predicted[0][0],
            "yes": predicted[0][1],
        }
        return r


class Bayes(object):
    def __init__(self, labels):
        self.bayes = {
            label: Baye()
            for label in labels
        }

    def reset(self):
        for v in self.bayes.values():
            v.reset()

    def fit(self, label, present, text):
        self.bayes[label].fit("yes" if present else "no", text)

    def seal(self):
        for v in self.bayes.values():
            v.seal()

    def score(self, label, text):
        score = self.bayes[label].score(text, label)
        return {
            "yes": score['yes'] if 'yes' in score else 0,
            "no": score['no'] if 'no' in score else 0,
        }


@functools.total_ordering
class Document(object):
    def __init__(self, core, doc_id, text, labels):
        self.core = core
        self.doc_id = doc_id
        self.text = text
        self.labels = labels
        self.scores = {}

    def __lt__(self, other):
        return self.doc_id < other.doc_id

    def __eq__(self, other):
        return self.doc_id == self.doc_id

    def __hash__(self):
        return hash(self.doc_id)

    def fit(self, bayes, all_labels):
        for label in all_labels:
            bayes.fit(label, label in self.labels, self.text)

    def compute_scores(self, bayes, all_labels):
        for label in all_labels:
            self.scores[label] = bayes.score(label, self.text)

    def compute_accuracy(self, all_labels):
        success = 0
        for label in all_labels:
            score = self.scores[label]
            guessed_has_label = (score['yes'] > score['no'])
            if guessed_has_label == (label in self.labels):
                success += 1
        return success / len(all_labels)


class Corpus(object):
    """
    All the texts of all the documents, and their labels.
    """
    def __init__(self, core):
        self.core = core
        self.all_documents = []
        self.documents = []
        self.labels = set()
        self.bayes = None

    def reset(self):
        self.bayes.reset()

    def load_all(self):
        self.all_documents = []
        docs = []
        self.core.call_all("storage_get_all_docs", docs)
        total = len(docs)
        for (idx, (doc_id, doc_url)) in enumerate(docs):
            self.core.call_all(
                "on_progress", "load_docs",
                idx / total, "Loading document %s" % doc_id
            )
            text = []
            self.core.call_all("doc_get_text_by_url", text, doc_url)
            text = "\n\n".join(text)

            labels = set()
            self.core.call_all("doc_get_labels_by_url", labels, doc_url)
            labels = [label[0] for label in labels]
            labels.sort()
            for label in labels:
                self.labels.add(label)

            self.all_documents.append(
                Document(self.core, doc_id, text, labels)
            )

        self.all_documents.sort()
        self.all_documents.reverse()
        self.core.call_all("on_progress", "load_docs", 1.0)

        self.bayes = Bayes(self.labels)

    def _get_doc_with_label(self, label, backlog):
        nb = 0
        for doc in self.all_documents:
            if label in doc.labels:
                yield doc
                nb += 1
                if nb >= backlog:
                    break

    def _get_doc_without_label(self, label, backlog):
        nb = 0
        for doc in self.all_documents:
            if label not in doc.labels:
                yield doc
                nb += 1
                if nb >= backlog:
                    break

    def fit(self, backlog):
        docs = set()
        for label in self.labels:
            docs.update(self._get_doc_with_label(label, backlog))
            docs.update(self._get_doc_without_label(label, backlog))

        self.documents = set(self.all_documents)
        for doc in docs:
            self.documents.remove(doc)
        self.documents = list(self.documents)
        self.documents.sort()
        self.documents.reverse()
        self.documents = self.documents[:200]
        if len(self.documents) < 200:
            print("Not enough documents left ({})".format(
                len(self.documents)
            ))
            return False

        for (idx, doc) in enumerate(docs):
            self.core.call_all(
                "on_progress", "training",
                idx / len(docs),
                "Extracting features from %d documents" % len(docs)
            )
            doc.fit(self.bayes, self.labels)
        self.core.call_all("on_progress", "training", 1.0)
        print("Extracted features from %d documents" % len(docs))
        return True

    def seal(self):
        print("Training...")
        self.bayes.seal()
        print("Training done")

    def compute_scores(self):

        total = len(self.documents)
        for (idx, doc) in enumerate(self.documents):
            self.core.call_all(
                "on_progress", "scoring",
                idx / total, "computing label scores on %s" % doc.doc_id
            )
            doc.compute_scores(self.bayes, self.labels)
        self.core.call_all("on_progress", "scoring", 1.0)

    def compute_accuracy(self):
        accuracy_sum = 0

        total = len(self.documents)
        for doc in self.documents:
            accuracy_sum += doc.compute_accuracy(self.labels)
        return accuracy_sum / total


class Plugin(openpaperwork_core.PluginBase):
    """
    Add a command to search for the best backlog for
    paperwork_backend.guesswork.label.sklearn
    heuristically and based on the user's documents.
    """
    def get_interfaces(self):
        return [
            'shell',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'doc_labels',
                'defaults': ['paperwork_backend.model.labels'],
            },
            {
                'interface': 'doc_text',
                'defaults': [
                    'paperwork_backend.model.hocr',
                    'paperwork_backend.model.pdf',
                ],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        parser.add_parser('compute_sklearn_label_guessing_backlog')

    def cmd_run(self, console, args):
        if args.command != 'compute_sklearn_label_guessing_backlog':
            return None

        corpus = Corpus(self.core)
        corpus.load_all()

        for backlog in (1, 10, 25, 50, 75, 100, 150, 200, 250, 300, 500, 2500):
            console.print("")
            console.print("Backlog {}:".format(backlog))
            corpus.reset()
            if not corpus.fit(backlog):
                return {}
            start = time.time()
            corpus.seal()
            stop = time.time()
            corpus.compute_scores()
            accuracy = corpus.compute_accuracy()
            console.print(
                "Backlog: {} ; Accuracy: {} ; Training time: {}s".format(
                    backlog, accuracy, int(stop - start)
                )
            )
        return {}



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2023  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import rich.text

import openpaperwork_core


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 1000

    def __init__(self):
        self.output = []
        self.console = None

    def get_interfaces(self):
        return ['print']

    def cmd_set_console(self, console):
        self.console = console

    def print(self, txt):
        self.output.append(txt)
        return True

    def print_isatty(self):
        if self.console is not None:
            return True
        return None

    def print_flush(self):
        if self.console is None:
            return None
        output = rich.text.Text("\n").join([
            rich.text.Text.from_ansi(line)
            for line in self.output
        ])
        self.console.console.print(output)
        self.output = []
        return True



import datetime
import logging
import time

import whoosh.fields
import whoosh.index
import whoosh.qparser
import whoosh.query
import whoosh.sorting

import openpaperwork_core

from .. import (_, sync)


LOGGER = logging.getLogger(__name__)
ID = "index"

WHOOSH_SCHEMA = whoosh.fields.Schema(
    docid=whoosh.fields.ID(stored=True, unique=True, sortable=True),
    docfilehash=whoosh.fields.ID(),
    content=whoosh.fields.TEXT(spelling=True),
    label=whoosh.fields.KEYWORD(commas=True, scorable=True),
    date=whoosh.fields.DATETIME(sortable=True),
    last_read=whoosh.fields.DATETIME(stored=True),
)


class CustomFuzzySearch(whoosh.qparser.query.FuzzyTerm):
    def __init__(
                self, fieldname, text, boost=1.0, maxdist=1,
                prefixlength=0, constantscore=True
            ):
        whoosh.qparser.query.FuzzyTerm.__init__(
            self, fieldname, text, boost, maxdist,
            prefixlength, constantscore=True
        )


class WhooshTransaction(sync.BaseTransaction):
    """
    Transaction to apply on the index. Methods may be slow but they
    are thread-safe.
    """
    def __init__(self, plugin, total_expected=-1):
        super().__init__(plugin.core, total_expected)

        self.priority = plugin.PRIORITY

        LOGGER.debug("Starting Whoosh index transaction")
        self.core = plugin.core
        self.writer = None
        self.modified = 0

        self.writer = plugin.index.writer()

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cancel()

    def __del__(self):
        self.cancel()

    def _update_doc_in_index(self, doc_id):
        """
        Collect infos on the document and add/update a document in the index
        """
        doc_url = self.core.call_success("doc_id_to_url", doc_id)

        doc_mtime = self.core.call_success("doc_get_mtime_by_url", doc_url)
        if doc_mtime is None:
            doc_mtime = 0
        doc_mtime = datetime.datetime.fromtimestamp(doc_mtime)

        doc_hash = self.core.call_success("doc_get_hash_by_url", doc_url)
        if doc_hash is None:
            # we get a hash only for PDF documents, not image documents.
            doc_hash = "undefined"
        else:
            doc_hash = ("%X" % doc_hash)

        doc_text = []
        self.core.call_all("doc_get_text_by_url", doc_text, doc_url)
        doc_text = "\n\n".join(doc_text)
        doc_text = self.core.call_success("i18n_strip_accents", doc_text)

        doc_labels = set()
        self.core.call_all("doc_get_labels_by_url", doc_labels, doc_url)
        doc_labels = ",".join([label[0] for label in doc_labels])
        doc_labels = self.core.call_success("i18n_strip_accents", doc_labels)

        doc_date = self.core.call_success("doc_get_date_by_id", doc_id)
        if doc_date is None:
            doc_date = datetime.datetime(year=1970, month=1, day=1)

        query = whoosh.query.Term("docid", doc_id)
        self.writer.delete_by_query(query)

        self.writer.update_document(
            docid=doc_id,
            docfilehash=doc_hash,
            content=doc_text,
            label=doc_labels,
            date=doc_date,
            last_read=doc_mtime
        )

    def add_doc(self, doc_id):
        LOGGER.info("Adding document '%s' to index", doc_id)
        self.notify_progress(
            ID, _("Indexing new document %s") % doc_id
        )
        self._update_doc_in_index(doc_id)
        self.modified += 1
        super().add_doc(doc_id)

    def del_doc(self, doc_id):
        LOGGER.info("Removing document '%s' from index", doc_id)
        self.notify_progress(
            ID, _("Removing document %s from index") % doc_id
        )
        query = whoosh.query.Term("docid", doc_id)
        self.writer.delete_by_query(query)
        self.modified += 1
        super().del_doc(doc_id)

    def upd_doc(self, doc_id):
        LOGGER.info("Updating document '%s' in index", doc_id)
        self.notify_progress(
            ID, _("Indexing updated document %s") % doc_id
        )
        self._update_doc_in_index(doc_id)
        self.modified += 1
        super().upd_doc(doc_id)

    def unchanged_doc(self, doc_id):
        self.notify_progress(
            ID, _("Examining document %s: unchanged") % (doc_id)
        )
        super().unchanged_doc(doc_id)

    def cancel(self):
        if self.writer is None:
            return

        self.core.call_one(
            "mainloop_schedule", self.core.call_all,
            'on_index_cancel'
        )
        LOGGER.info("Canceling transaction")
        self.writer.cancel()
        self.writer = None
        self.core.call_one(
            "mainloop_schedule", self.core.call_all,
            "on_index_updated"
        )
        self.notify_done(ID)

    def commit(self):
        self.notify_progress(
            ID, _("Committing changes in the index ...")
        )
        self.core.call_one(
            "mainloop_schedule", self.core.call_all,
            'on_index_commit_start'
        )
        if self.modified <= 0:
            LOGGER.info(
                "commit() called but nothing to commit."
                " Cancelling transaction"
            )
            self.writer.cancel()
            self.writer = None
        else:
            LOGGER.info(
                "Committing %d changes to Whoosh index", self.modified
            )
            self.writer.commit()
            self.writer = None
        self.notify_done(ID)
        self.core.call_success(
            "mainloop_schedule", self.core.call_all,
            'on_index_commit_end'
        )


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.obs_callbacks = []
        self.query_parsers = {
            'strict': [],
            'fuzzy': [],
        }
        self.index = None

        self.local_dir = None
        self.index_dir = None

    def get_interfaces(self):
        return [
            "index",
            "suggestions",
            "syncable",
        ]

    def get_deps(self):
        return [
            {
                'interface': 'data_versioning',
                'defaults': ['openpaperwork_core.data_versioning'],
            },
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'data_dir_handler',
                'defaults': ['paperwork_backend.datadirhandler'],
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
            # Optional dependencies:
            # {
            #     'interface': 'page_boxes',
            #     'defaults': [
            #         'paperwork_backend.model.hocr',
            #         'paperwork_backend.model.pdf',
            #     ],
            # },
            # {
            #     'interface': 'doc_hash',
            #     'defaults': ['paperwork_backend.model.pdf'],
            # },
            # {
            #     'interface': 'doc_labels',
            #     'defaults': ['paperwork_backend.model.labels'],
            # },
        ]

    def init(self, core):
        super().init(core)
        self._init()

    def _init(self):
        data_dir = self.core.call_success(
            "data_dir_handler_get_individual_data_dir"
        )
        self.index_dir = self.core.call_success(
            "fs_join", data_dir, "index"
        )

        need_index_rewrite = True
        while need_index_rewrite:
            try:
                LOGGER.info(
                    "Opening Whoosh index '%s' ...", self.index_dir
                )
                self.index = whoosh.index.open_dir(
                    self.core.call_success("fs_unsafe", self.index_dir)
                )
                # check that the schema is up-to-date
                # We use the string representation of the schemas, because
                # previous versions of whoosh don't always implement __eq__
                if str(self.index.schema) != str(WHOOSH_SCHEMA):
                    raise Exception("Index version mismatch")
                need_index_rewrite = False
            except Exception as exc:
                LOGGER.warning(
                    "Failed to open index '%s': %s."
                    " Will rebuild index from scratch",
                    self.index_dir, exc
                )

            if need_index_rewrite:
                self._destroy()

                LOGGER.info("Creating a new index")

                self.core.call_success("fs_mkdir_p", self.index_dir)

                new_index = whoosh.index.create_in(
                    self.core.call_success("fs_unsafe", self.index_dir),
                    WHOOSH_SCHEMA
                )
                new_index.close()
                LOGGER.info("Index '%s' created", self.index_dir)

        self.query_parsers = {
            'fuzzy': [
                whoosh.qparser.MultifieldParser(
                    ["label", "content"], schema=self.index.schema,
                    termclass=CustomFuzzySearch
                ),
                whoosh.qparser.MultifieldParser(
                    ["label", "content"], schema=self.index.schema,
                    termclass=whoosh.qparser.query.Prefix
                ),
            ],
            'strict': [
                whoosh.qparser.MultifieldParser(
                    ["label", "content"], schema=self.index.schema,
                    termclass=whoosh.query.Term
                ),
            ],
        }

    def on_data_dir_changed(self):
        self._close()
        self._init()

    def _close(self):
        LOGGER.info("Closing Whoosh index")
        if self.index is not None:
            self.index.close()
        self.index = None

    def _destroy(self):
        self._close()
        if self.core.call_success("fs_exists", self.index_dir) is not None:
            LOGGER.warning("Destroying the index ...")
            self.core.call_success("fs_rm_rf", self.index_dir)
            LOGGER.warning("Index destroyed")

    def doc_transaction_start(self, out: list, total_expected=-1):
        out.append(WhooshTransaction(self, total_expected))

    def index_search(self, out: list, query, limit=None, search_type='fuzzy'):
        start = time.time()

        out_set = set()

        query = query.strip()
        query = self.core.call_success("i18n_strip_accents", query)
        if query == "":
            queries = [whoosh.query.Every()]
        else:
            queries = []
            for parser in self.query_parsers[search_type]:
                queries.append(parser.parse(query))

        with self.index.searcher() as searcher:
            for q in queries:
                facet = whoosh.sorting.FieldFacet("docid", reverse=True)
                results = searcher.search(q, limit=limit, sortedby=facet)
                has_results = False
                for result in results:
                    has_results = True
                    doc_id = result['docid']
                    doc_url = self.core.call_success("doc_id_to_url", doc_id)
                    if doc_url is None:
                        continue
                    out_set.add((doc_id, doc_url))
                    if limit is not None and len(out_set) >= limit:
                        break
                if has_results:
                    break

        out += out_set

        stop = time.time()
        LOGGER.info(
            "Search [%s] took %dms (limit=%s, type=%s)",
            query, (stop - start) * 1000, limit, search_type
        )

    def index_get_doc_id_by_hash(self, doc_hash):
        doc_hash = "%X" % doc_hash
        with self.index.searcher() as searcher:
            results = searcher.search(
                whoosh.query.Term('docfilehash', doc_hash)
            )
            if len(results) <= 0:
                return None
            return results[0]

    def suggestion_get(self, out: set, sentence):
        query_parser = self.query_parsers['strict'][0]
        query = query_parser.parse(sentence)

        with self.index.searcher() as searcher:
            corrected = searcher.correct_query(
                query, sentence, correctors={
                    'content': searcher.corrector("content"),
                    'label': searcher.corrector("label"),
                }
            )
            if corrected.query != query:
                out.add(corrected.string)

    def sync(self, promises: list):
        """
        Requests the document list from the document storage plugin
        and updates the index accordingly.

        This call is asynchronous and use the main loop to do its job.
        """
        storage_all_docs = []

        promise = openpaperwork_core.promise.ThreadedPromise(
            self.core, self.core.call_all,
            args=("storage_get_all_docs", storage_all_docs,)
        )
        promise = promise.then(lambda *args, **kwargs: None)

        class IndexDoc(object):
            def __init__(s, index_result):
                (s.key, s.extra) = index_result

        def get_index_docs():
            with self.index.searcher() as searcher:
                index_all_docs = searcher.search(
                    whoosh.query.Every(), limit=None
                )
                index_all_docs = [
                    (result['docid'], result['last_read'])
                    for result in index_all_docs
                ]
                index_all_docs = [IndexDoc(r) for r in index_all_docs]
            return index_all_docs

        promise = promise.then(
            lambda: (
                [
                    sync.StorageDoc(self.core, doc[0], doc[1])
                    for doc in storage_all_docs
                ],
                get_index_docs()
            )
        )
        promise = promise.then(
            lambda args: (
                args[0], args[1],
                [WhooshTransaction(
                    self, max(len(storage_all_docs), len(args[1]))
                )]
            )
        )
        promise = promise.then(lambda args: sync.Syncer(
            self.core, ["whoosh"], args[0], args[1], args[2]
        ))
        promise = promise.then(openpaperwork_core.promise.ThreadedPromise(
            self.core, lambda syncer: syncer.run()
        ))
        promises.append(promise)



import logging
import os
import subprocess
import tempfile

import openpaperwork_core
import paperwork_backend.deps

from .. import _


LOGGER = logging.getLogger(__name__)


LIBREOFFICE = [
    r'libreoffice',
    r'soffice',
    r'soffice.exe',
    r'C:\Program Files\LibreOffice\program\soffice.exe',
    r'/app/libreoffice/program/soffice.bin',  # Flatpak
]

LIBREOFFICE_ARGS = [
    '--nocrashreport',
    '--nodefault',
    '--nofirststartwizard',
    '--nolockcheck',
    '--nologo',
    '--norestore',
    '--headless',
    '--convert-to', 'pdf',
    '--outdir', '{out_dir}',
    '{in_doc}',
]


def is_exe(fpath):
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def which(program):
    if os.path.sep in program:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None


class Plugin(openpaperwork_core.PluginBase):
    FILE_TYPES = [
        (
            "application/msword", "doc",
            "Microsoft Word (.doc)",
        ),
        (
            "application/msword", "dot",
            _("Microsoft Word template (.dot)"),
        ),
        (
            "application/vnd.ms-excel", "xls",
            "Microsoft Excel (.xls)",
        ),
        (
            "application/vnd.ms-excel", "xlt",
            _("Microsoft Excel template (.xlt)"),
        ),
        (
            "application/vnd.ms-powerpoint", "pps",
            "Microsoft PowerPoint (.pps)",
        ),
        (
            "application/vnd.ms-powerpoint", "ppt",
            _("Microsoft PowerPoint template (.ppt)"),
        ),
        (
            "application/vnd.oasis.opendocument.chart", "odc",
            _("OpenOffice/LibreOffice Chart (.odc)"),
        ),
        (
            "application/vnd.oasis.opendocument.database", "odb",
            _("OpenOffice/LibreOffice Database (.odb)"),
        ),
        (
            "application/vnd.oasis.opendocument.formula", "odf",
            _("OpenOffice/LibreOffice Formula (.odf)"),
        ),
        (
            "application/vnd.oasis.opendocument.graphics", "odg",
            _("OpenOffice/LibreOffice Graphics (.odg)"),
        ),
        (
            "application/vnd.oasis.opendocument.graphics-template", "otg",
            _("OpenOffice/LibreOffice Graphics template (.otg)"),
        ),
        (
            "application/vnd.oasis.opendocument.image", "odi",
            _("OpenOffice/LibreOffice Image template (.odi)"),
        ),
        (
            "application/vnd.oasis.opendocument.presentation", "odp",
            _("OpenOffice/LibreOffice Presentation (.odp)"),
        ),
        (
            "application/vnd.oasis.opendocument.presentation-template", "otp",
            _("OpenOffice/LibreOffice Presentation template (.otp)"),
        ),
        (
            "application/vnd.oasis.opendocument.spreadsheet", "ods",
            _("OpenOffice/LibreOffice Spreadsheet (.ods)"),
        ),
        (
            "application/vnd.oasis.opendocument.spreadsheet-template", "ots",
            _("OpenOffice/LibreOffice Spreadsheet template (.ots)"),
        ),
        (
            "application/vnd.oasis.opendocument.text", "odt",
            _("OpenOffice/LibreOffice Text (.odt)"),
        ),
        (
            "application/vnd.oasis.opendocument.text-master", "odm",
            _("OpenOffice/LibreOffice Text master (.odm)"),
        ),
        (
            "application/vnd.oasis.opendocument.text-template", "ott",
            _("OpenOffice/LibreOffice Text template (.ott)"),
        ),
        (
            "application/vnd.oasis.opendocument.text-web", "oth",
            _("OpenOffice/LibreOffice Text web (.oth)"),
        ),
        (
            "application/vnd.openxmlformats-officedocument.presentationml"
            ".presentation",
            "pptx",
            "Microsoft PowerPoint presentation (.pptx)",
        ),
        (
            "application/vnd.openxmlformats-officedocument.presentationml"
            ".slide",
            "sldx",
            _("Microsoft PowerPoint slide (.sldx)"),
        ),
        (
            "application/vnd.openxmlformats-officedocument.presentationml"
            ".slideshow",
            "ppsx",
            _("Microsoft PowerPoint slideshow (.ppsx)"),
        ),
        (
            "application/vnd.openxmlformats-officedocument.presentationml"
            ".template",
            "potx",
            _("Microsoft PowerPoint presentation template (.potx)"),
        ),
        (
            "application/vnd.openxmlformats-officedocument.spreadsheetml"
            ".sheet",
            "xlsx",
            "Microsoft Excel (.xlsx)",
        ),
        (
            "application/vnd.openxmlformats-officedocument.spreadsheetml"
            ".template",
            "xltx",
            _("Microsoft Excel template (.xltx)"),
        ),
        (
            "application/vnd.openxmlformats-officedocument"
            ".wordprocessingml.document",
            "docx",
            "Microsoft Word (.docx)",
        ),
        (
            "application/vnd.openxmlformats-officedocument"
            ".wordprocessingml.template",
            "dotx",
            _("Microsoft Word template (.dotx)"),
        ),
    ]

    def __init__(self):
        self.libreoffice = None

    def get_interfaces(self):
        return [
            "chkdeps",
            "doc_converter",
        ]

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio']
            },
        ]

    def init(self, core):
        super().init(core)
        for exe in LIBREOFFICE:
            self.libreoffice = which(exe)
            if self.libreoffice is not None:
                break
        LOGGER.info("Libreoffice: %s", self.libreoffice)

    def chkdeps(self, out: dict):
        if self.libreoffice is None:
            out['libreoffice'] = paperwork_backend.deps.LIBREOFFICE

    def converter_get_file_types(self, out: list):
        if self.libreoffice is None:
            return
        out += self.FILE_TYPES

    def convert_file_to_pdf(self, doc_file_uri, mime_type, out_pdf_file_url):
        if self.libreoffice is None:
            return None

        for (mime, ext, desc) in self.FILE_TYPES:
            if mime_type == mime:
                break
        else:
            return None

        LOGGER.info(
            "Converting %s (%s) to %s (PDF)",
            doc_file_uri, mime_type, out_pdf_file_url
        )

        file_name = self.core.call_success("fs_basename", doc_file_uri)
        if "." not in file_name:
            LOGGER.error("No file extension ? %s", doc_file_uri)
            return None
        file_ext = file_name.rsplit(".", 1)[-1]

        cwd = os.getcwd()
        with tempfile.TemporaryDirectory() as tmp_dir:
            src_file = os.path.join(tmp_dir, "doc." + file_ext)
            dst_file = os.path.join(tmp_dir, "doc.pdf")

            # Assume Libreoffice only supports local files
            self.core.call_success(
                "fs_copy", doc_file_uri,
                self.core.call_success("fs_safe", src_file)
            )
            if not self.core.call_success(
                    "fs_exists", self.core.call_success("fs_safe", src_file)):
                LOGGER.error("Failed to copy file %s", doc_file_uri)
                return None

            os.chdir(tmp_dir)
            try:
                args = [
                    x.format(in_doc=src_file, out_dir=tmp_dir)
                    for x in LIBREOFFICE_ARGS
                ]
                popen = subprocess.Popen(
                    [self.libreoffice] + args,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE
                )
                with popen:
                    (stdout, stderr) = popen.communicate()

                dst_file_url = self.core.call_success("fs_safe", dst_file)

                if not self.core.call_success("fs_exists", dst_file_url):
                    LOGGER.error(
                        "Failed to convert %s (%s) to %s|%s (PDF)",
                        doc_file_uri, mime_type, dst_file_url, out_pdf_file_url
                    )
                    LOGGER.error("Command was: %s", [self.libreoffice] + args)
                    LOGGER.error("LibreOffice stdout: %s", stdout)
                    LOGGER.error("LibreOffice stderr: %s", stderr)
                    return False

                self.core.call_success(
                    "fs_copy",
                    dst_file_url,
                    out_pdf_file_url
                )
            finally:
                os.chdir(cwd)
        return True



"""
To use it:

```sh
paperwork-cli plugins add \
    paperwork_backend.guesswork.label.sklearn.compute_threshold
paperwork-cli compute_sklearn_label_guessing_threshold
```
"""

import functools
import itertools
import time

import sklearn.naive_bayes
import sklearn.feature_extraction.text

import openpaperwork_core


NB_ITERATIONS = 2  # beware complexity will increase exponentionnally


def pairwise(iterable):
    (a, b) = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


class Baye(object):
    samples = 5

    def __init__(self):
        self.data = []
        self.targets = []

        self.sklearn_count_vectorizer = \
            sklearn.feature_extraction.text.CountVectorizer()
        self.sklearn_tfid_transformer = \
            sklearn.feature_extraction.text.TfidfTransformer(use_idf=False)

        # For reference, always returning "no" gives an accuracy of ~0.9475
        # For reference, always returning "yes" gives an accuracy of ~0.0524
        # For reference, simplebayes: accuracy=~0.9912 with threshold ~0.162

        # Best accuracy: ~0.9991 ; No threshold usable / to use
        self.sklearn_classifier = sklearn.naive_bayes.GaussianNB()

        # Best accuracy: ~0.9634 ; Best threshold: ~0.068
        # self.sklearn_classifier = sklearn.naive_bayes.MultinomialNB()

        # Best accuracy: ~0.9642 ; Best threshold: ~0.078
        # self.sklearn_classifier = sklearn.naive_bayes.ComplementNB()

        # Best accuracy: ~0.9687; Best threshold: ~0.9980
        # self.sklearn_classifier = sklearn.naive_bayes.BernoulliNB()

        # Slow as hell.
        # self.sklearn_classifier = sklearn.naive_bayes.CategoricalNB()

    def fit(self, category, text):
        text = text.strip()
        if len(text) <= 0:
            return

        category = (1 if category == 'yes' else 0)
        self.data.append(text)
        self.targets.append(category)

    def seal(self):
        counts = self.sklearn_count_vectorizer.fit_transform(self.data)
        tfid = self.sklearn_tfid_transformer.fit_transform(counts)
        self.sklearn_classifier.fit(tfid.toarray(), self.targets)

    def score(self, text, label):
        counts = self.sklearn_count_vectorizer.transform([text])
        tfid = self.sklearn_tfid_transformer.transform(counts)
        predicted = self.sklearn_classifier.predict_proba(tfid.toarray())
        r = {
            "no": predicted[0][0],
            "yes": predicted[0][1],
        }
        if self.samples > 0:
            self.samples -= 1
            print("Prediction sample: {}={}".format(label, r))
        return r


class Bayes(object):
    def __init__(self, labels):
        self.bayes = {
            label: Baye()
            for label in labels
        }

    def fit(self, label, present, text):
        self.bayes[label].fit("yes" if present else "no", text)

    def seal(self):
        for v in self.bayes.values():
            v.seal()

    def score(self, label, text):
        score = self.bayes[label].score(text, label)
        return {
            "yes": score['yes'] if 'yes' in score else 0,
            "no": score['no'] if 'no' in score else 0,
        }


@functools.total_ordering
class Document(object):
    def __init__(self, core, doc_id, text, labels):
        self.core = core
        self.doc_id = doc_id
        self.text = text
        self.labels = labels
        self.scores = {}

    def __lt__(self, other):
        return self.doc_id < other.doc_id

    def __eq__(self, other):
        return self.doc_id == self.doc_id

    def __hash__(self):
        return hash(self.doc_id)

    def fit(self, bayes, all_labels):
        for label in all_labels:
            bayes.fit(label, label in self.labels, self.text)

    def compute_scores(self, bayes, all_labels):
        for label in all_labels:
            self.scores[label] = bayes.score(label, self.text)

    def compute_accuracy(self, threshold, all_labels):
        success = 0
        for label in all_labels:
            score = self.scores[label]
            total = score['yes'] + score['no']
            if total == 0:
                score = 0
            else:
                score = score['yes'] / total
            guessed_has_label = score > threshold
            if guessed_has_label == (label in self.labels):
                success += 1
        return success / len(all_labels)


class Corpus(object):
    """
    All the texts of all the documents, and their labels.
    """
    def __init__(self, core):
        self.core = core
        self.documents = []
        self.labels = set()
        self.bayes = None

    def load_all(self):
        docs = []
        self.core.call_all("storage_get_all_docs", docs)
        total = len(docs)
        for (idx, (doc_id, doc_url)) in enumerate(docs):
            self.core.call_all(
                "on_progress", "load_docs",
                idx / total, "Loading document %s" % doc_id
            )
            text = []
            self.core.call_all("doc_get_text_by_url", text, doc_url)
            text = "\n\n".join(text)

            labels = set()
            self.core.call_all("doc_get_labels_by_url", labels, doc_url)
            labels = [label[0] for label in labels]
            labels.sort()
            for label in labels:
                self.labels.add(label)

            self.documents.append(Document(self.core, doc_id, text, labels))

        self.documents.sort()
        self.core.call_all("on_progress", "load_docs", 1.0)

        self.bayes = Bayes(self.labels)

    def fit(self):
        for (idx, doc) in enumerate(self.documents):
            doc.fit(self.bayes, self.labels)

    def seal(self):
        self.bayes.seal()

    def compute_scores(self):
        total = len(self.documents)
        for (idx, doc) in enumerate(self.documents):
            self.core.call_all(
                "on_progress", "scoring",
                idx / total, "computing label scores on %s" % doc.doc_id
            )
            doc.compute_scores(self.bayes, self.labels)
        self.core.call_all("on_progress", "scoring", 1.0)

    def compute_accuracy(self, threshold):
        accuracy_sum = 0

        total = len(self.documents)
        for doc in self.documents:
            accuracy_sum += doc.compute_accuracy(threshold, self.labels)
        return accuracy_sum / total


class Plugin(openpaperwork_core.PluginBase):
    """
    Add a command to search for the best threshold for
    paperwork_backend.guesswork.label.simplebayes (see THRESHOLD_YES_NO_RATIO)
    heuristically and based on the user's documents.

    OBSOLETE: Now that we are using sklearn.GaussianNB, there is no need for
    a threshold anymore.
    """
    def get_interfaces(self):
        return [
            'shell',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'doc_labels',
                'defaults': ['paperwork_backend.model.labels'],
            },
            {
                'interface': 'doc_text',
                'defaults': [
                    'paperwork_backend.model.hocr',
                    'paperwork_backend.model.pdf',
                ],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        parser.add_parser('compute_sklearn_label_guessing_threshold')

    @staticmethod
    def _get_next_thresholds(thresholds):
        """
        Heuristic figuring out the next thresholds to try to estimate
        """
        middles = []
        for (threshold, next_threshold) in pairwise(thresholds):
            middles.append(
                (
                    (threshold[0] + next_threshold[0]) / 2,
                    (threshold[1] + next_threshold[1]) / 2,
                )
            )
        return [m[0] for m in middles]

    def cmd_run(self, console, args):
        if args.command != 'compute_sklearn_label_guessing_threshold':
            return None

        corpus = Corpus(self.core)
        corpus.load_all()
        corpus.fit()
        corpus.seal()
        corpus.compute_scores()

        iterations = NB_ITERATIONS
        thresholds = [
            (threshold, corpus.compute_accuracy(threshold))
            for threshold in (0.0, 1.0)
        ]

        for iteration in range(0, iterations):
            console.print("")
            best = max(thresholds, key=lambda x: x[1])
            console.print(
                "Iteration {}: best accuracy={}, best threshold={}".format(
                    iteration, best[1], best[0]
                )
            )

            start = time.time()
            next_thresholds = self._get_next_thresholds(thresholds)
            total = len(next_thresholds)
            for (idx, next_threshold) in enumerate(next_thresholds):
                self.core.call_all(
                    "on_progress", "iteration",
                    idx / total, "Iteration %d" % iteration
                )
                thresholds.append(
                    (next_threshold, corpus.compute_accuracy(next_threshold)),
                )
            thresholds.sort()
            self.core.call_all("on_progress", "iteration", 1.0)
            stop = time.time()

            s = int(stop - start)
            ms = int(((stop - start) * 1000) % 1000)
            console.print("Iteration took {}s {}ms".format(s, ms))

        console.print("")
        for threshold in thresholds:
            console.print("Threshold: {}".format(threshold))

        console.print("")
        best = max(thresholds, key=lambda x: x[1])
        console.print("RESULT: best accuraccy={}, best threshold={}".format(
            best[1], best[0]
        ))

        return {
            "all": thresholds,
            "best": best,
        }



"""
Provides support for URIs "file://".
"""

import codecs
import logging
import os
import os.path
import shutil
import tempfile

from . import CommonFsPluginBase


LOGGER = logging.getLogger(__name__)


class Plugin(CommonFsPluginBase):
    def __init__(self):
        super().__init__()
        self.tmp_files = set()

    def _uri_to_path(self, uri):
        if not uri.lower().startswith("file://"):
            return None
        return self.fs_unsafe(uri)

    def fs_open(self, uri, mode='r', **kwargs):
        path = self._uri_to_path(uri)
        if path is None:
            return None

        if 'w' not in mode and 'a' not in mode and not os.path.exists(path):
            return None

        if 'b' in mode:
            return open(path, mode)
        return codecs.open(path, mode, encoding='utf-8')

    def fs_exists(self, uri):
        path = self._uri_to_path(uri)
        if path is None:
            return None

        r = os.path.exists(path)
        if not r:
            return None
        return r

    def fs_listdir(self, uri):
        path = self._uri_to_path(uri)
        if path is None:
            return None
        if not os.path.exists(path):
            return None
        for f in os.listdir(path):
            yield self.fs_safe(os.path.join(path, f))

    def fs_rename(self, old_uri, new_uri):
        old_path = self._uri_to_path(old_uri)
        new_path = self._uri_to_path(new_uri)
        if old_path is None or new_path is None:
            return None
        if not os.path.exists(old_path):
            return None

        os.rename(old_path, new_path)
        return True

    def fs_unlink(self, uri, **kwargs):
        path = self._uri_to_path(uri)
        if path is None:
            return

        if not os.path.exists(path):
            return None
        os.unlink(path)
        return True

    def fs_rm_rf(self, uri, **kwargs):
        path = self._uri_to_path(uri)
        if path is None:
            return

        shutil.rmtree(path, ignore_errors=True)
        return True

    def fs_get_mtime(self, uri):
        path = self._uri_to_path(uri)
        if path is None:
            return

        return int(os.stat(path).st_mtime)

    def fs_getsize(self, uri):
        path = self._uri_to_path(uri)
        if path is None:
            return

        return os.stat(path).st_size

    def fs_isdir(self, uri):
        path = self._uri_to_path(uri)
        if path is None:
            return

        if os.path.isdir(path):
            return True
        return None

    def fs_copy(self, old_uri, new_uri):
        old_path = self._uri_to_path(old_uri)
        new_path = self._uri_to_path(new_uri)
        if old_path is None or new_path is None:
            return None

        shutil.copy(old_path, new_path)
        return True

    def fs_mkdir_p(self, uri):
        path = self._uri_to_path(uri)
        if path is None:
            return

        os.makedirs(path, mode=0o700, exist_ok=True)
        return True

    def fs_recurse(self, parent_uri, dir_included=False):
        path = self._uri_to_path(parent_uri)
        if path is None:
            return

        for (root, dirs, files) in os.walk(path):
            if dir_included:
                for d in dirs:
                    p = os.path.join(path, root, d)
                    yield self.fs_safe(p)
            for f in files:
                p = os.path.join(path, root, f)
                yield self.fs_safe(p)

    def fs_hide(self, uri):
        pass

    def fs_get_mime(self, uri):
        path = self._uri_to_path(uri)
        if path is None:
            return None

        # should use 'magic', but Core can't have any dependency on it.
        # other we would pull it on all platforms.
        path = path.lower()
        if path.endswith(".pdf"):
            return "application/pdf"
        if path.endswith(".png"):
            return "image/png"
        if path.endswith(".tiff"):
            return "image/tiff"
        if path.endswith(".bmp"):
            return "image/x-ms-bmp"
        if path.endswith(".jpeg") or path.endswith(".jpg"):
            return "image/jpeg"
        if path.endswith(".txt"):
            return "text/plain"

        return None

    def fs_iswritable(self, uri):
        path = self._uri_to_path(uri)
        if path is None:
            return None

        return os.access(path, os.W_OK)

    def fs_mktemp(
                self, prefix=None, suffix=None, mode='w+b', on_disk=False,
                **kwargs
            ):
        if 'b' not in mode:
            tmp = tempfile.NamedTemporaryFile(
                prefix=prefix, suffix=suffix, delete=False, mode=mode,
                encoding='utf-8'
            )
        else:
            tmp = tempfile.NamedTemporaryFile(
                prefix=prefix, suffix=suffix, delete=False, mode=mode
            )
        self.tmp_files.add(tmp.name)
        return (self.fs_safe(tmp.name), tmp)

    def on_quit(self):
        for tmp_file in self.tmp_files:
            try:
                os.unlink(tmp_file)
            except FileNotFoundError:
                pass



import io
import logging

import PIL
import PIL.Image

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)


def surface2image(core, surface):
    """
    Convert a cairo surface into a PIL image
    """
    # XXX(Jflesch): Python 3 problem
    # cairo.ImageSurface.get_data() raises NotImplementedYet ...

    # import PIL.ImageDraw
    #
    # if surface is None:
    #     return None
    # dimension = (surface.get_width(), surface.get_height())
    # img = PIL.Image.frombuffer("RGBA", dimension,
    #                            surface.get_data(), "raw", "BGRA", 0, 1)
    #
    # background = PIL.Image.new("RGB", img.size, (255, 255, 255))
    # background.paste(img, mask=img.split()[3])  # 3 is the alpha channel
    # return background

    core.call_all("on_perfcheck_start", "surface2image")

    img_io = io.BytesIO()
    surface.surface.write_to_png(img_io)
    img_io.seek(0)
    try:
        img = PIL.Image.open(img_io)
    except PIL.Image.DecompressionBombError as exc:
        LOGGER.warning("Cairo surface is too big", exc_info=exc)
        return core.call_success("pillow_get_error", "too_big")
    core.call_all("on_objref_track", img)
    img.load()

    if "A" not in img.getbands():
        core.call_all("on_perfcheck_stop", "surface2image", size=img.size)
        return img

    img_no_alpha = PIL.Image.new("RGB", img.size, (255, 255, 255))
    core.call_all("on_objref_track", img_no_alpha)
    img_no_alpha.paste(img, mask=img.split()[3])  # 3 is the alpha channel
    core.call_all(
        "on_perfcheck_stop", "surface2image", size=img_no_alpha.size
    )
    return img_no_alpha


class Plugin(openpaperwork_core.PluginBase):
    FILE_EXTENSION = ".pdf"

    def get_interfaces(self):
        return [
            'page_img_size',
            'pillow',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'pdf_cairo_url',
                'defaults': ['paperwork_backend.cairo.poppler'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'pillow_util',
                'defaults': ['openpaperwork_core.pillow.util'],
            },
            {
                'interface': 'urls',
                'defaults': ['openpaperwork_core.urls'],
            },
        ]

    def _check_is_pdf(self, file_url):
        (url, args) = self.core.call_success("url_args_split", file_url)
        if not url.lower().endswith(self.FILE_EXTENSION):
            return (None, None, None)

        page_idx = int(args.get("page", 1)) - 1
        password = args.get('password', None)
        if password is not None:
            password = bytes.fromhex(password).decode("utf-8")

        return (file_url, page_idx, password)

    def url_to_pillow(self, file_url):
        (file_url, page_idx, password) = self._check_is_pdf(file_url)
        if file_url is None:
            return None

        pillow = self.core.call_one(  # Poppler is not really thread safe
            "mainloop_execute", self._url_to_pillow,
            file_url, page_idx, password
        )
        return pillow

    def cairo_surface_to_pillow(self, surface):
        return surface2image(self.core, surface)

    def _url_to_pillow(self, file_url, page_idx, password):
        surface = self.core.call_success(
            "pdf_page_to_cairo_surface", file_url, page_idx, password
        )
        # WORKAROUND(Jflesch): may occur with corrupted PDFs
        if surface is None:
            return None

        img = surface2image(self.core, surface)
        surface.surface.finish()
        img.load()
        return img

    def url_to_pillow_promise(self, file_url):
        (doc_url, page_idx, password) = self._check_is_pdf(file_url)
        if doc_url is None:
            return None

        return openpaperwork_core.promise.Promise(
            self.core, self.url_to_pillow, args=(file_url,)
        )

    def pillow_to_url(self, *args, **kwargs):
        # It could be implemented, but there are no known use-cases.
        return None



#!/usr/bin/python3

import logging
import time

import PIL.Image

import openpaperwork_core
import openpaperwork_core.promise

from . import util


THUMBNAIL_WIDTH = 64
THUMBNAIL_HEIGHT = 80

PAGE_THUMBNAIL_FILENAME = 'paper.{}.thumb.jpg'

LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 5000  # see page_get_img_url()

    def get_interfaces(self):
        return [
            'pages',
            'thumbnail',  # thumbnail_get_path()
            'thumbnailer',  # thumbnail_from_img()
        ]

    def get_deps(self):
        return [
            {
                'interface': 'page_img',
                'defaults': [
                    'paperwork_backend.model.img',
                    'paperwork_backend.model.pdf',
                ],
            },
            {
                'interface': 'pillow',
                'defaults': [
                    'openpaperwork_core.pillow.img',
                    'paperwork_backend.pillow.pdf',
                ],
            },
            {
                'interface': 'pillow_util',
                'defaults': ['openpaperwork_core.pillow.util'],
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
        ]

    def thumbnail_get_doc(self, doc_url):
        return self.thumbnail_get_page(doc_url, page_idx=0)

    def thumbnail_get_doc_promise(self, doc_url):
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self.thumbnail_get_doc, args=(doc_url,)
        )

    def thumbnail_from_img(self, img):
        (width, height) = img.size
        scale = max(
            float(width) / THUMBNAIL_WIDTH,
            float(height) / THUMBNAIL_HEIGHT
        )
        width /= scale
        height /= scale
        width = max(width, 1)
        height = max(height, 1)
        return img.resize(
            (int(width), int(height)),
            getattr(PIL.Image, 'Resampling', PIL.Image).LANCZOS
        )

    def thumbnail_get_page(self, doc_url, page_idx):
        thumbnail_url = self.core.call_success(
            "fs_join", doc_url, PAGE_THUMBNAIL_FILENAME.format(page_idx + 1)
        )
        page_url = self.core.call_success(
            "page_get_img_url", doc_url, page_idx
        )
        if page_url is None:
            LOGGER.warning(
                "Failed to get thumbnail for %s p%d. No page URL",
                doc_url, page_idx
            )
            return None

        if self.core.call_success("fs_exists", thumbnail_url) is not None:
            thumbnail_mtime = self.core.call_success(
                "fs_get_mtime", thumbnail_url
            )
            if not self.core.call_success("fs_iswritable", thumbnail_url):
                page_mtime = self.core.call_success("fs_get_mtime", page_url)
                if thumbnail_mtime < page_mtime:
                    self.core.call_success("fs_unlink", thumbnail_url)

        if self.core.call_success("fs_exists", thumbnail_url) is not None:
            LOGGER.debug("Loading thumbnail for %s page %d", doc_url, page_idx)
            try:
                thumbnail = self.core.call_success(
                    "url_to_pillow", thumbnail_url
                )
                size = thumbnail.size
                if size[0] == THUMBNAIL_WIDTH or size[1] == THUMBNAIL_HEIGHT:
                    return thumbnail
                LOGGER.info(
                    "Thumbnail for %s page %d doesn't have the expected size"
                    " (%s instead of %s). Regenerating.",
                    doc_url, page_idx, size,
                    (THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT)
                )
            except Exception as exc:
                LOGGER.warning(
                    "Failed to load thumbnail for %s page %d. Regenerating",
                    doc_url, page_idx, exc_info=exc
                )

        LOGGER.info("Generating thumbnail for %s page %d", doc_url, page_idx)
        start = time.time()
        try:
            page = self.core.call_success("url_to_pillow", page_url)
        except Exception as exc:
            LOGGER.error("Failed to load page %s", page_url, exc_info=exc)
            page = self.core.call_success("pillow_get_error", "invalid")

        thumbnail = self.thumbnail_from_img(page)

        self.core.call_success(
            "pillow_to_url", thumbnail, thumbnail_url,
            format='JPEG', quality=0.85
        )

        stop = time.time()
        LOGGER.info(
            "Thumbnail for %s page %d generated in %f seconds",
            doc_url, page_idx, stop - start
        )
        return thumbnail

    def thumbnail_get_page_promise(self, doc_url, page_idx):
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self.thumbnail_get_page, args=(doc_url, page_idx)
        )

    def page_delete_by_url(self, doc_url, page_idx):
        return util.delete_page_file(
            self.core, PAGE_THUMBNAIL_FILENAME, doc_url, page_idx,
            trash=False
        )

    def page_move_by_url(
                self,
                source_doc_url, source_page_idx,
                dest_doc_url, dest_page_idx
            ):
        return util.move_page_file(
            self.core, PAGE_THUMBNAIL_FILENAME,
            source_doc_url, source_page_idx,
            dest_doc_url, dest_page_idx
        )

    def page_reset_by_url(self, doc_url, page_idx):
        # see model.img_overlay.page_reset_by_url()
        # We must force a rebuild of the thumbnail
        self.page_delete_by_url(doc_url, page_idx)

    def page_get_img_url(self, doc_url, page_idx, write=False, **kwargs):
        if not write:
            return None
        # Image is going to be updated --> We must force a rebuild of the
        # thumbnail
        self.page_delete_by_url(doc_url, page_idx)
        return None



"""
Spatial indexer that works in a way very similar to R-tree. Except for the
way the nodes are split because I'm a lazy bastard and it's good enough.
"""
import math
import logging

import openpaperwork_core


LOGGER = logging.getLogger(__name__)
NB_CHILDREN_PER_NODES = 4  # must be a multiple of 2 because I'm a lazy bastard

# Node: R-tree page (leaf or not)
# Non-leaf: children are other nodes
# Leaf: last R-tree node: children are tuples (box, obj)


def compute_area(box):
    return abs((box[1][0] - box[0][0]) * (box[1][1] - box[0][1]))


class Node(object):
    def __init__(self, box=None):
        self.leaf = True
        self.children = []

        self.box = None
        self.area = 0
        self.set_box(box)

        self.parent = None

    def compute_enlargment(self, box):
        enlarged = (
            (
                min(self.box[0][0], box[0][0]),
                min(self.box[0][1], box[0][1]),
            ),
            (
                max(self.box[1][0], box[1][0]),
                max(self.box[1][1], box[1][1]),
            ),
        )
        enlarged = compute_area(enlarged)
        return enlarged - self.area

    def set_box(self, box):
        self.box = box
        if box is None:  # root node
            self.area = math.inf
        else:
            self.area = compute_area(box)

    def recompute_box(self, recurse=True):
        depth = 0
        s = self
        while True:
            depth += 1
            assert depth < 512
            if s.leaf:
                box = (
                    (
                        min(x[0][0][0] for x in s.children),
                        min(x[0][0][1] for x in s.children),
                    ),
                    (
                        max(x[0][1][0] for x in s.children),
                        max(x[0][1][1] for x in s.children),
                    ),
                )
            else:
                box = (
                    (
                        min(x.box[0][0] for x in s.children),
                        min(x.box[0][1] for x in s.children),
                    ),
                    (
                        max(x.box[1][0] for x in s.children),
                        max(x.box[1][1] for x in s.children),
                    ),
                )
            s.set_box(box)
            if not recurse or s.parent is None:
                break
            s = s.parent

    def is_full(self):
        assert self.leaf
        return len(self.children) >= NB_CHILDREN_PER_NODES

    @staticmethod
    def _pick_seed(boxes):
        """
        Returns a pair of nodes that would be the most wasteful.
        (see Guttman Quadratic Split algorithm)
        """

        assert len(boxes) >= 2

        if len(boxes) == 2:  # minor optim
            return tuple(boxes)

        max_waste = -1
        r = None

        for (a_idx, a) in enumerate(boxes):
            for b in boxes[a_idx + 1:]:
                new_rect = (
                    (
                        min(a[0][0][0], b[0][0][0]),
                        min(a[0][0][1], b[0][0][1]),
                    ),
                    (
                        max(a[0][1][0], b[0][1][0]),
                        max(a[0][1][1], b[0][1][1]),
                    )
                )
                new_area = (
                    (new_rect[1][0] - new_rect[0][0]) *
                    (new_rect[1][1] - new_rect[0][1])
                )
                waste = new_area - a[2] - b[2]
                if waste > max_waste or r is None:
                    r = (a, b)

        assert r is not None
        return r

    def split(self):
        assert self.leaf

        our_children = self.children
        picked = []

        while len(our_children) > 0:
            p = self._pick_seed(our_children)
            picked.append(p)
            for c in p:
                our_children.remove(c)

        node_a = Node()
        node_a.parent = self
        node_a.children = [b[0] for b in picked]
        node_a.recompute_box(recurse=False)

        node_b = Node()
        node_b.parent = self
        node_b.children = [b[1] for b in picked]
        node_b.recompute_box(recurse=False)

        self.children = [node_a, node_b]

        self.leaf = False

    def insert_box(self, box):
        assert self.leaf
        self.children.append(box)
        self.recompute_box()

    def __gt__(self, o):
        return False


class RTreeSpatialIndexer(object):
    def __init__(self, boxes):
        self.root = Node()

        boxes = list(boxes)
        LOGGER.info("Loading %d boxes in rtree", len(boxes))

        for (pos, obj) in boxes:
            self.insert(pos, obj)

        # self.print_stats()

    def print_stats(self):
        """
        Not used at the moment, but useful to make sure the tree looks OK when
        debugging.
        """
        nb_boxes = 0
        nb_nodes = 0
        max_depth = 0

        to_examine = [(0, self.root)]
        while len(to_examine) > 0:
            nb_nodes += 1

            (depth, n) = to_examine.pop(0)
            max_depth = max(max_depth, depth)

            if n.leaf:
                nb_boxes += len(n.children)
            else:
                for c in n.children:
                    to_examine.append((depth + 1, c))

        LOGGER.info(
            "Rtree: %d boxes, %d nodes, max depth: %d",
            nb_boxes, nb_nodes, max_depth
        )

    def _choose_leaf(self, box_position, node):
        # keep picking up the child node that overlap as much of
        # box_position as possible
        return min({
            (c.compute_enlargment(box_position), c)
            for c in node.children
        })[1]

    def _find_leaf(self, box_position, root=None):
        if root is None:
            n = self.root
        else:
            n = root
        while not n.leaf:
            n = self._choose_leaf(box_position, n)
        return n

    def insert(self, box_position, obj):
        new_child = (box_position, obj, compute_area(box_position))
        leaf = self._find_leaf(box_position)
        if leaf.is_full():
            leaf.split()
            leaf = self._find_leaf(box_position, leaf)
        leaf.insert_box(new_child)

    @staticmethod
    def is_in(box_position, pt_x, pt_y):
        if box_position[0][0] > pt_x:
            return False
        if box_position[0][1] > pt_y:
            return False
        if box_position[1][0] < pt_x:
            return False
        if box_position[1][1] < pt_y:
            return False
        return True

    def get_boxes(self, pt_x, pt_y):
        """
        Returns all the boxes that contains the point (pt_x, pt_y)
        """
        examined = 0
        to_examine = [self.root]
        while len(to_examine) > 0:
            n = to_examine.pop(0)
            examined += 1
            if not n.leaf:
                for c in n.children:
                    if self.is_in(c.box, pt_x, pt_y):
                        to_examine.append(c)
            else:
                for c in n.children:
                    if self.is_in(c[0], pt_x, pt_y):
                        yield (c[0], c[1])
        # LOGGER.debug("Examined nodes: %d", examined)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'spatial_index'
        ]

    def spatial_indexer_get(self, boxes):
        return RTreeSpatialIndexer(boxes)



import logging

import openpaperwork_core
import openpaperwork_core.deps


GI_AVAILABLE = False
GDK_AVAILABLE = False

try:
    import gi
    GI_AVAILABLE = True
except (ImportError, ValueError):
    pass

if GI_AVAILABLE:
    try:
        gi.require_version('Gdk', '3.0')
        from gi.repository import Gdk
        GDK_AVAILABLE = True
    except (ImportError, ValueError):
        pass


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.refcount = 0
        self.windows = []
        self.realize_handler_id = None

    def get_interfaces(self):
        return [
            'busy',
            'chkdeps',
            'gtk_window_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_mainwindow',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GDK_AVAILABLE:
            out['gdk'].update(openpaperwork_core.deps.GDK)

    def _set_mouse_cursor(self, offset):
        self.refcount += offset
        assert self.refcount >= 0

        if len(self.windows) <= 0:
            LOGGER.warning(
                "Cannot change mouse cursor: no main window defined"
            )
            return

        if self.refcount > 0:
            LOGGER.info("Mouse cursor --> busy")
            try:
                display = self.windows[-1].get_display()
                cursor = Gdk.Cursor.new_for_display(
                    display, Gdk.CursorType.WATCH
                )
            except TypeError as exc:
                # may happen with Wayland
                LOGGER.error("Failed to switch mouse cursor", exc_info=exc)
                return
        else:
            LOGGER.info("Mouse cursor --> idle")
            cursor = None
        self.windows[-1].get_window().set_cursor(cursor)

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def on_busy(self):
        self._set_mouse_cursor(1)

    def on_idle(self):
        self._set_mouse_cursor(-1)

    def on_widget_busyness_changed(self, widget, busy):
        if busy:
            cursor = None
        else:
            display = widget.get_display()
            cursor = Gdk.Cursor.new_for_display(display, Gdk.CursorType.WATCH)
        window = widget.get_window()

        def _set_cursor(*args, **kwargs):
            window = widget.get_window()
            window.set_cursor(cursor)
            if self.realize_handler_id is not None:
                widget.disconnect(self.realize_handler_id)
                self.realize_handler_id = None

        if window is not None:
            _set_cursor()
        else:
            # assuming the widget will be realized soon
            widget.connect("realize", _set_cursor)



import logging

import openpaperwork_core

from .. import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    """
    Locates (and optionally deletes) files at the root of the work directory.
    """
    def get_interfaces(self):
        return ['chkworkdir']

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'doc_pdf_import',
                'defaults': ['paperwork_backend.model.pdf'],
            },
        ]

    def check_work_dir(self, out_problems: list):
        all_docs = []
        self.core.call_all("storage_get_all_docs", all_docs, only_valid=False)
        all_docs.sort()

        total = len(all_docs)
        LOGGER.info("Checking work directory (%d documents)", total)
        for (idx, (doc_id, doc_url)) in enumerate(all_docs):
            self.core.call_all(
                "on_progress", "chkworkdir_file_at_workdir_root",
                idx / total, _("Checking doc %s") % (doc_id,)
            )
            isdir = self.core.call_success("fs_isdir", doc_url)
            if isdir:
                continue
            if doc_url.lower().endswith(".pdf"):
                out_problems.append({
                    "problem": "file_at_workdir_root",
                    "doc_id": doc_id,
                    "doc_url": doc_url,
                    "human_description": {
                        "problem": _("Document %s is not a directory") % (
                            doc_id,
                        ),
                        "solution": _(
                            "Turn %s into <YYYYMMDD_hhmm_ss>/doc.pdf"
                        ) % (
                            doc_id,
                        ),
                    },
                    "solution": "import_pdf",
                })
            else:
                out_problems.append({
                    "problem": "file_at_workdir_root",
                    "doc_id": doc_id,
                    "doc_url": doc_url,
                    "human_description": {
                        "problem": _("Document %s is not a directory") % (
                            doc_id,
                        ),
                        "solution": _("Delete document %s") % (doc_id,),
                    },
                    "solution": "delete",
                })

        self.core.call_all(
            "on_progress", "chkworkdir_file_at_workdir_root", 1.0
        )

    def fix_work_dir(self, problems):
        total = len(problems)
        for (idx, problem) in enumerate(problems):
            if problem['problem'] != 'file_at_workdir_root':
                continue
            self.core.call_all(
                "on_progress", "fixworkdir_file_at_workdir_root",
                idx / total,
                _("Fixing %s") % (problem['doc_url'],)
            )
            if problem['solution'] == "import_pdf":
                LOGGER.info("Importing %s", problem['doc_url'])
                self.core.call_success("doc_pdf_import", problem['doc_url'])
                self.core.call_success("fs_unlink", problem['doc_url'])
            else:
                LOGGER.info("Deleting %s", problem['doc_url'])
            self.core.call_all("fs_unlink", problem['doc_url'])
        self.core.call_all(
            "on_progress", "fixworkdir_file_at_workdir_root", 1.0
        )



import os
import multiprocessing
import platform
import psutil
import sys

try:
    import distro
except (ImportError, ValueError):
    assert os.name == "nt"

import openpaperwork_core


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 1000

    def get_interfaces(self):
        return ['stats']

    def stats_get(self, out: dict):
        if os.name == 'nt':
            distribution = str(platform.win32_ver())
        else:
            distribution = str((
                distro.id(),
                distro.version(),
                distro.name()
            ))
        processor = ""
        os_name = os.name
        if os_name != 'nt':  # processor contains too much infos on Windows
            processor = str(platform.processor())

        cpu_freq = None
        if hasattr(psutil, 'cpu_freq'):
            cpu_freq = psutil.cpu_freq()
        if cpu_freq is not None:
            cpu_freq = int(cpu_freq.max)

        if cpu_freq is not None:
            out['cpu_freq'] = cpu_freq
        out['cpu_count'] = multiprocessing.cpu_count()
        out['os_name'] = os_name
        out['platform_architecture'] = str(platform.architecture())
        out['platform_distribution'] = distribution
        out['platform_machine'] = platform.machine()
        out['platform_mem'] = int(psutil.virtual_memory().total)
        out['platform_processor'] = processor
        out['software_python'] = sys.version
        out['software_release'] = platform.release()
        out['software_system'] = platform.system()
        return out



import itertools
import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)
IDX_GENERATOR = itertools.count()


class PageSelectionHandler(object):
    def __init__(self, core, plugin, page):
        self.core = core
        self.plugin = plugin
        self.page = page

        self.actives = []
        self.boxes = None

        self.orig = (-1, -1)
        self.first = None
        self.last = None

        self.gesture = Gtk.GestureDrag.new(page.widget)
        self.gesture.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)

        self.signal_handlers = []

    def set_boxes(self, line_boxes, spatial_index):
        self.boxes = spatial_index

    def connect(self):
        self.disconnect()
        handlers = (
            ('drag-begin', self.on_drag_begin),
            ('drag-update', self.on_drag_update),
            ('drag-end', self.on_drag_end),
        )
        for (signal, cb) in handlers:
            self.signal_handlers.append(
                (signal, self.gesture.connect(signal, cb))
            )

    def disconnect(self):
        for (signal, handler_id) in self.signal_handlers:
            self.gesture.disconnect(handler_id)
        self.signal_handlers = []

    def _find_box(self, x, y):
        if self.boxes is None:
            return None
        zoom = self.page.get_zoom()
        x /= zoom
        y /= zoom
        nboxes = self.boxes.get_boxes(x, y)

        # take the box with the smallest area
        try:
            nbox = min({
                (
                    abs(
                        (n[1].box.position[1][0] - n[1].box.position[0][0]) *
                        (n[1].box.position[1][1] - n[1].box.position[0][1])
                    ), n[1]
                )
                for n in nboxes
            })[1]
        except ValueError:
            return None
        return nbox

    def _get_selected(self):
        if self.first is None:
            return []

        if self.first.index <= self.last.index:
            first = self.first
            last = self.last
        else:
            first = self.last
            last = self.first

        current = first

        while current is not None and current != last:
            yield current.box
            current = current.next

        yield last.box

    def on_drag_begin(self, gesture, x, y):
        if not self.plugin.enabled:
            self.first = None
            self.last = None
            return

        self.actives = []
        self.orig = (x, y)
        self.first = self._find_box(x, y)
        self.last = self.first

    def on_drag_update(self, gesture, x, y):
        if self.first is None:
            return

        (x, y) = (self.orig[0] + x, self.orig[1] + y)
        last = self._find_box(x, y)

        if last is None or last == self.last:
            return

        self.last = last
        self.page.widget.queue_draw()

        LOGGER.info(
            "Text selection: first: %d ; last: %d",
            self.first.index, self.last.index
        )

    def on_drag_end(self, gesture, x, y):
        if self.first is None:
            return

        self.on_drag_update(gesture, x, y)
        self.core.call_all(
            "on_page_boxes_selected",
            self.page.doc_id, self.page.doc_url, self.page.page_idx,
            list(self._get_selected())
        )

    def draw(self, cairo_ctx):
        if not self.plugin.enabled:
            return

        for box in self._get_selected():
            self.core.call_all(
                "page_draw_box",
                cairo_ctx, self.page,
                box.position, (0.0, 1.0, 0.0),
                border_width=2
            )


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10

    def __init__(self):
        super().__init__()
        self.handlers = {}
        self.enabled = False

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_pageview_boxes_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_pageview',
                'defaults': ['paperwork_gtk.mainwindow.docview.pageview'],
            },
            {
                'interface': 'gtk_pageview_boxes',
                'defaults': [
                    'paperwork_gtk.mainwindow.docview.pageview.boxes'
                ],
            },
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_docview_closed_page(self, page):
        self.handlers.pop(page.widget).disconnect()

    def on_page_boxes_loaded(self, page, boxes, spatial_index):
        h = PageSelectionHandler(self.core, self, page)
        self.handlers[page.widget] = h

        assert page.widget in self.handlers
        h = self.handlers[page.widget]
        h.set_boxes(boxes, spatial_index)
        if page.get_visible():
            h.connect()

    def on_page_visibility_changed(self, page, visible):
        if page.widget not in self.handlers:
            return
        h = self.handlers[page.widget]
        if visible:
            h.connect()
        else:
            h.disconnect()

    def on_page_draw(self, cairo_ctx, page):
        if page.widget not in self.handlers:
            return
        self.handlers[page.widget].draw(cairo_ctx)

    def on_layout_change(self, layout_name):
        self.enabled = (layout_name == 'paged')



import sys

from . import PluginBase


class Plugin(PluginBase):
    def __init__(self):
        super().__init__()
        self.original_hook = sys.excepthook

    def get_interfaces(self):
        return ['uncaught_exception']

    def get_deps(self):
        return [
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
        ]

    def init(self, core):
        super().init(core)
        sys.excepthook = self._on_uncaught_exception

    def _on_uncaught_exception(self, exc_type, exc_value, exc_tb):
        exc_info = (exc_type, exc_value, exc_tb)
        try:
            self.core.call_one(
                "mainloop_execute", self._broadcast_exception, exc_info
            )
        finally:
            if getattr(sys, 'frozen', False):
                # Assumes that cx_freeze has put a specific handler
                # for uncatched exceptions (popup and stuff)
                self.original_hook(exc_type, exc_value, exc_tb)

    def _broadcast_exception(self, exc_info):
        # make sure we don't loop
        sys.excepthook = self.original_hook
        try:
            nb = self.core.call_all("on_uncaught_exception", exc_info)
            if nb <= 0:
                # no log handler yet --> switch back to default
                self.original_hook(*exc_info)
        finally:
            sys.excepthook = self._on_uncaught_exception



import atexit
import code
import os
import os.path
import readline
import rlcompleter  # noqa: F401
import time
import threading

from . import PluginBase


class HistoryConsole(code.InteractiveConsole):
    def __init__(
            self, locals=None, filename="<console>",
            histfile=os.path.expanduser("~/.console-history")):
        super().__init__(locals, filename)
        self.init_history(histfile)

    def init_history(self, histfile):
        readline.parse_and_bind("tab: complete")
        if hasattr(readline, "read_history_file"):
            try:
                readline.read_history_file(histfile)
            except FileNotFoundError:
                pass
            atexit.register(self.save_history, histfile)

    def save_history(self, histfile):
        readline.set_history_length(1000)
        readline.write_history_file(histfile)


class ProxyCore(object):
    def __init__(self, core):
        self.core = core

    def call_all(self, *args, **kwargs):
        return self.core.call_one(
            "mainloop_execute", self.core.call_all, *args, **kwargs
        )

    def call_success(self, *args, **kwargs):
        return self.core.call_one(
            "mainloop_execute", self.core.call_success, *args, **kwargs
        )

    def call_one(self, *args, **kwargs):
        return self.core.call_one(
            "mainloop_execute", self.core.call_one, *args, **kwargs
        )


class Plugin(PluginBase):
    def __init__(self):
        super().__init__()
        self.has_quit = False

        # used for wait():
        self.progress_condition = threading.Condition()
        self.progresses = {}
        self._previous_progresses = {}

        self.nb_windows_to_realize = 0

    def get_interfaces(self):
        return [
            'gtk_window_listener'
        ]

    def get_deps(self):
        return [
            {
                'interface': 'data_versioning',
                'defaults': ['openpaperwork_core.data_versioning'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
            {
                'interface': 'paths',
                'defaults': ['openpaperwork_core.paths.xdg'],
            },
        ]

    def init(self, core):
        super().init(core)
        data_dir = self.core.call_success("paths_get_data_dir")
        base_hist_dir = self.core.call_success(
            "fs_join", data_dir, "openpaperwork"
        )
        self.core.call_success("fs_mkdir_p", base_hist_dir)
        histfile = self.core.call_success(
            "fs_join", base_hist_dir, "interactive_history"
        )

        print("Objects provided:")
        print("  core : reference to OpenPaperwork's core")
        print("  stop(): shut down the application (but not this shell)")
        print("  exit(): stops this shell (but not the application)")
        print("  wait(): wait for all background tasks to end")
        print("  Ctrl-D / EOF: stops this shell and the application")

        console = HistoryConsole({
            "core": ProxyCore(core),
            "stop": self._stop,
            "wait": self._wait,
        }, histfile=self.core.call_success("fs_unsafe", histfile))
        threading.Thread(target=self._interact, args=(console,)).start()

    def on_gtk_window_opened(self, window):
        with self.progress_condition:
            if window.get_window() is not None:
                return

            self.nb_windows_to_realize += 1
            realize_handler_id = None

            def on_realize():
                self.nb_windows_to_realize -= 1
                window.disconnect(realize_handler_id)
                self.progress_condition.notify_all()

            realize_handler_id = window.connect("realize", on_realize)

    def on_gtk_window_closed(self, window):
        pass

    def _interact(self, console):
        console.interact()
        self._stop()

    def on_quit(self):
        self.has_quit = True

    def _stop(self):
        if self.has_quit:
            return
        print("Quitting")
        self.core.call_one(
            "mainloop_execute", self.core.call_all, "mainloop_quit_graceful"
        )

    def on_progress(self, upd_type, progress, description=None):
        progress = int(progress * 100)
        with self.progress_condition:
            if progress >= 100:
                if upd_type in self.progresses:
                    self.progresses.pop(upd_type)
                    self.progress_condition.notify_all()
            elif (upd_type not in self.progresses or
                    progress != self.progresses[upd_type]):
                self.progresses[upd_type] = progress
                self.progress_condition.notify_all()

    def _wait(self):
        # WORKAROUND(Jflesch): sometimes, for some reason, we never get the
        # notification for the end of boxes loading
        MAX_TIME = 30  # seconds

        start = time.time()
        time_diff = 0

        # wait a little, in case the call to _wait() came slightly before
        # the background task creation.
        print("Waiting for all background tasks to end")
        time.sleep(3.0)
        with self.progress_condition:
            while ((len(self.progresses) > 0 or
                    self.nb_windows_to_realize > 0) and
                    time_diff <= MAX_TIME):
                while ((len(self.progresses) > 0 or
                        self.nb_windows_to_realize > 0) and
                        time_diff <= MAX_TIME):
                    print("Waiting for all background tasks to end")
                    print("Remaining background tasks:")
                    for (upd_type, progress) in self.progresses.items():
                        print("  {}: {}%".format(upd_type, int(progress)))
                    self.progress_condition.wait(1.0)
                    time_diff = time.time() - start
                if time_diff > MAX_TIME:
                    break
                # wait again a little ; sometime background tasks are
                # sequentially removed and added
                self.progress_condition.release()
                try:
                    time.sleep(3)
                finally:
                    self.progress_condition.acquire()
        if time_diff <= MAX_TIME:
            print("All background tasks have ended")
        else:
            print("TIMEOUT")



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False
try:
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_gtk
import openpaperwork_gtk.deps

from ... import _


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "page_move_inside_doc"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.active_doc = None
        self.active_page_idx = -1
        self.windows = []

    def get_interfaces(self):
        return [
            'action',
            'action_page',
            'action_page_move_inside_doc',
            'chkdeps',
            'doc_open',
            'gtk_window_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'pages',
                'defaults': [
                    'paperwork_backend.model.img',
                    'paperwork_backend.model.img_overlay',
                    'paperwork_backend.model.hocr',
                    'paperwork_backend.model.pdf',
                ],
            },
            {
                'interface': 'gtk_dialog_single_entry',
                'defaults': ['openpaperwork_gtk.dialogs.single_entry'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GLIB_AVAILABLE:
            return

        action = Gio.SimpleAction.new(ACTION_NAME, None)
        action.connect("activate", self._start_move_inside_doc)
        self.core.call_all("app_actions_add", action)

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def doc_open(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)

    def doc_close(self):
        self.active_doc = None

    def on_page_shown(self, page_idx):
        self.active_page_idx = page_idx

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def _show_error(self, msg):
        flags = (
            Gtk.DialogFlags.MODAL |
            Gtk.DialogFlags.DESTROY_WITH_PARENT
        )
        dialog = Gtk.MessageDialog(
            transient_for=self.windows[-1],
            flags=flags,
            message_type=Gtk.MessageType.ERROR,
            buttons=Gtk.ButtonsType.OK,
            text=msg
        )
        dialog.connect("response", lambda dialog, response: dialog.destroy())
        dialog.show_all()

    def _start_move_inside_doc(self, *args, **kwargs):
        assert self.active_doc is not None
        assert self.active_page_idx is not None
        msg = _("Move the page %d to what position ?") % (
            self.active_page_idx + 1
        )
        self.core.call_success(
            "gtk_show_dialog_single_entry",
            self, msg, str(self.active_page_idx + 1),
            active_doc=self.active_doc,
            active_page_idx=self.active_page_idx
        )

    def on_dialog_single_entry_reply(
            self, origin, r, new_position, *args, **kwargs):
        if origin is not self:
            return None
        if not r:
            return None
        try:
            new_position = int(new_position) - 1
        except ValueError:
            self._show_error(_("Invalid page position: %s") % (new_position))
            return True

        active_doc = kwargs['active_doc']
        active_page_idx = kwargs['active_page_idx']

        nb_pages = self.core.call_success(
            "doc_get_nb_pages_by_url", active_doc[1]
        )
        if new_position >= nb_pages or new_position < 0:
            self._show_error(
                _(
                    "Invalid page position: %d."
                    " Out of document bounds (1-%d)."
                ) % (
                    new_position,
                    nb_pages
                )
            )
            return True
        if new_position == active_page_idx:
            self._show_error(_("Page position unchanged"))
            return True

        promise = openpaperwork_core.promise.Promise(self.core)
        promise = promise.then(
            self.core.call_all,
            "page_move_by_url",
            active_doc[1], active_page_idx,
            active_doc[1], new_position
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(
            self.core.call_all, "doc_reload", *active_doc
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(self.core.call_success(
            "transaction_simple_promise",
            [('upd', active_doc[0])]
        ))
        self.core.call_success("transaction_schedule", promise)



"""
Plugin providing a controller object for page editing. This controller object
uses a paperwork_backend.imgedit.AbstractPageEditorUI object to tells the UI
what to do. The UI must reciprocate by transmitting some events to the
controller object.
"""

import logging

import openpaperwork_core
import openpaperwork_core.promise

from .. import _


LOGGER = logging.getLogger(__name__)


class Frame(object):
    def __init__(self, editor, original):
        # convert the frame coordinates into a more convenient format
        self.editor = editor
        self.original = original

    def get(self):
        frame = None
        for (e, img_size) in zip(
                    self.editor.active_modifiers, self.editor.img_sizes
                ):
            if hasattr(e, 'frame'):
                frame = e.frame
            elif frame is not None:
                frame = e.transform_frame(img_size, frame)
        return frame

    def set(self, frame):
        for (e, img_size) in zip(
                    reversed(self.editor.active_modifiers),
                    reversed(self.editor.img_sizes)
                ):
            if hasattr(e, 'frame'):
                e.frame = frame
            else:
                frame = e.untransform_frame(img_size, frame)


class PageEditor(object):
    def __init__(self, core, doc_url, page_idx, page_editor_ui):
        self.core = core
        self.ui = page_editor_ui
        self.doc_url = doc_url
        self.page_idx = page_idx

        if doc_url is not None:
            page_url = self.core.call_success(
                "page_get_img_url", doc_url, page_idx, write=False
            )
            self.original_img = self.core.call_success(
                "url_to_pillow", page_url
            )

            # The frame coordinates we keep here are relative to the original
            # image (not the resulting image, that can, for instance, be
            # rotated)
            # TODO(Jflesch): We should use libpillowfight.scan_border()
            # to predefined an useful frame.
            original_frame = ((0, 0), self.original_img.size)
        else:
            self.original_img = None
            original_frame = ((0, 0), (0, 0))

        self.frame = Frame(self, original_frame)
        modifiers = []
        self.core.call_all("img_editor_get_names", modifiers)

        self.modifier_descriptors = {}
        if 'color_equalization' in modifiers:
            self.modifier_descriptors['color_equalization'] = {
                "id": "color_equalization",
                "name": _("Color equalization"),
                "modifier": 'color_equalization',
                "default_kwargs": {},
                "need_frame": False,
                "togglable": True,
                "enabled": False,
                "priority": -999,
            }
        if ('cropping' in modifiers
                and self.ui.can(self.ui.CAPABILITY_SHOW_FRAME)):
            self.modifier_descriptors['crop'] = {
                "id": "crop",
                "name": _("Cropping"),
                "modifier": 'cropping',
                "default_kwargs": {},
                "need_frame": True,
                "togglable": True,
                "enabled": False,
                "priority": 100,
            }
        if 'rotation' in modifiers:
            self.modifier_descriptors['rotate_clockwise'] = {
                "id": "rotate_clockwise",
                "name": _("Clockwise Rotation"),
                "modifier": 'rotation',
                "default_kwargs": {'angle': 90},
                "need_frame": False,
                "togglable": False,
                "priority": 50,
            }
            self.modifier_descriptors['rotate_counterclockwise'] = {
                "id": "rotate_counterclockwise",
                "name": _("Counterclockwise Rotation"),
                "modifier": 'rotation',
                "default_kwargs": {'angle': -90},
                "need_frame": False,
                "togglable": False,
                "priority": 49,
            }
        self.active_modifiers = []
        # image sizes before transformation of each modifier
        self.img_sizes = []

        if self.ui is not None:
            self._refresh_preview()
            self._refresh_frame()

    def get_modifiers(self):
        r = list(self.modifier_descriptors.values())
        r.sort(key=lambda m: -m['priority'])
        return r

    def _needs_frame(self):
        if not self.ui.can(self.ui.CAPABILITY_SHOW_FRAME):
            return
        for e in self.active_modifiers:
            if hasattr(e, 'frame'):
                return True
        return False

    def _refresh_modifiers(self):
        for modifier in self.modifier_descriptors.values():
            enabled = False
            if modifier['togglable'] and modifier['enabled']:
                enabled = True
            self.ui.set_modifier_state(modifier['id'], enabled)

    def _refresh_preview(self):
        img = self.original_img
        self.img_sizes = []
        for e in self.active_modifiers:
            self.img_sizes.append(img.size)
            img = e.transform(img, preview=True)
        self.core.call_one("mainloop_execute", self.ui.show_preview, img)

    def _refresh_frame(self):
        if not self._needs_frame():
            self.core.call_one("mainloop_execute", self.ui.hide_frame_selector)
            return
        self.core.call_one(
            "mainloop_execute", self.ui.show_frame_selector
        )

    def _on_modifier_selected(self, modifier_id):
        modifier_descriptor = self.modifier_descriptors[modifier_id]

        if modifier_descriptor['togglable'] and modifier_descriptor['enabled']:
            self.core.call_all(
                "img_editor_unset",
                self.active_modifiers,
                modifier_descriptor['modifier']
            )
            modifier_descriptor['enabled'] = False
        else:
            self.core.call_all(
                "img_editor_set",
                self.active_modifiers,
                modifier_descriptor['modifier'],
                **modifier_descriptor['default_kwargs']
            )
            modifier_descriptor['enabled'] = True

        self._refresh_modifiers()
        self._refresh_preview()
        self._refresh_frame()

    def on_modifier_selected(self, modifier_id):
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self._on_modifier_selected, args=(modifier_id,)
        )

    def _on_save(self):
        img = self.original_img
        for e in self.active_modifiers:
            img = e.transform(img, preview=False)
        page_url = self.core.call_success(
            "page_get_img_url", self.doc_url, self.page_idx, write=True
        )
        self.core.call_success("pillow_to_url", img, page_url)
        # Drop the text so the OCR will be run again
        self.core.call_success(
            "page_set_boxes_by_url", self.doc_url, self.page_idx, []
        )
        self.core.call_success(
            "mainloop_schedule", self.ui.on_edit_end,
            self.doc_url, self.page_idx
        )

    def on_save(self):
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self._on_save
        )

    def on_cancel(self):
        self.active_modifiers = []
        self.ui.on_edit_end(self.doc_url, self.page_idx)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['page_editor']

    def get_deps(self):
        return [
            {
                "interface": "img_editor",
                'defaults': [
                    'paperwork_backend.imgedit.color',
                    'paperwork_backend.imgedit.crop',
                    'paperwork_backend.imgedit.rotate',
                ],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'pillow',
                'defaults': [
                    'openpaperwork_core.pillow.img',
                    'paperwork_backend.pillow.pdf',
                ]
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
        ]

    def page_editor_get(self, doc_url, page_idx, page_editor_ui):
        return PageEditor(self.core, doc_url, page_idx, page_editor_ui)



import logging

import openpaperwork_core

from . import BaseDocViewController


LOGGER = logging.getLogger(__name__)


class ClickController(BaseDocViewController):
    def on_page_activated(self, page):
        super().on_page_activated(page)
        LOGGER.info("User activated page %d", page.page_idx)
        self.plugin.core.call_all("docview_set_layout", "paged")
        self.plugin.core.call_all("doc_goto_page", page.page_idx)

    def docview_set_layout(self, name):
        if name == 'grid':
            return
        self.plugin.docview_switch_controller('click', NoClickController)


class NoClickController(BaseDocViewController):
    def docview_set_layout(self, name):
        if name != 'grid':
            return
        self.plugin.docview_switch_controller('click', ClickController)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['gtk_docview_controller']

    def get_deps(self):
        return [
            {
                'interface': 'gtk_docview',
                'defaults': ['paperwork_gtk.mainwindow.docview'],
            },
        ]

    def gtk_docview_get_controllers(self, out: dict, docview):
        out['click'] = ClickController(docview)



import logging

import PIL
import PIL.Image
import PIL.ImageDraw
import PIL.ImageFont

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    ID_TO_MSG = {
        'invalid': "Invalid image",
        'too_big': "Image too big.\nImage not loaded.",
    }

    def get_interfaces(self):
        return ['pillow_util']

    def pillow_get_error(self, error_id):
        msg = self.ID_TO_MSG[error_id]
        font = PIL.ImageFont.load_default()

        longest_line = max(((len(line), line) for line in msg.split("\n")))
        longest_line = longest_line[1]
        if hasattr(font, 'getbbox'):
            size = font.getbbox(longest_line)
            size = (size[2], size[3])
        elif hasattr(font, 'getsize'):
            size = font.getsize(longest_line)
        else:
            size = (100, 100)
        size = (size[0], size[1] * (1 + (2 * msg.count("\n"))))

        img = PIL.Image.new("RGB", size, (255, 255, 255))
        draw = PIL.ImageDraw.Draw(img)
        draw.text((0, 0), msg, font=font, fill=(0, 0, 0))

        img = img.resize(
            tuple([s * 8 for s in size]),
            getattr(PIL.Image, 'Resampling', PIL.Image).LANCZOS
        )

        return img

    def pillow_add_border(self, img, color="#a6a5a4", width=1):
        """
        Add a border of the specified color and width around a PIL image
        """
        img_draw = PIL.ImageDraw.Draw(img)
        for line in range(0, width):
            img_draw.rectangle(
                [
                    (line, line),
                    (
                        img.size[0] - 1 - line,
                        img.size[1] - 1 - line
                    )
                ],
                outline=color
            )
        return img



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import openpaperwork_core

from .. import _


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                "interface": "extra_text",
                "defaults": ['paperwork_backend.model.extra_text'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        extra_text_parser = parser.add_parser(
            'extra_text', help=_(
                "Manage additional text attached to documents"
            )
        )

        subparser = extra_text_parser.add_subparsers(
            help=_('sub-command'), dest='subcommand', required=True
        )

        parser = subparser.add_parser(
            'get', help=_("Get a document additional text")
        )
        parser.add_argument('doc_id')

        parser = subparser.add_parser(
            'set', help=_("Set a document additional text")
        )
        parser.add_argument('doc_id')
        parser.add_argument('text')

    def cmd_run(self, console, args):
        if args.command != 'extra_text':
            return None
        doc_id = args.doc_id
        doc_url = self.core.call_success("doc_id_to_url", doc_id)
        if args.subcommand == "get":
            return self._cmd_get(console, doc_id, doc_url)
        elif args.subcommand == "set":
            return self._cmd_set(console, doc_id, doc_url, args.text)
        else:
            return None

    def _cmd_get(self, console, doc_id, doc_url):
        text = []
        self.core.call_success("doc_get_extra_text_by_url", text, doc_url)
        if len(text) > 0:
            console.print("  " + _("Additional text:"))
            console.print("\n".join(text))
        else:
            console.print(_("No additional text"))
        return text

    def _cmd_set(self, console, doc_id, doc_url, text):
        text = text.strip()
        self.core.call_success("doc_set_extra_text_by_url", doc_url, text)
        console.print(_("Done"))
        return True



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps

from ... import _


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "page_delete"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.active_doc = None
        self.active_page_idx = -1

    def get_interfaces(self):
        return [
            'action',
            'action_page',
            'action_page_delete',
            'chkdeps',
            'doc_open',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'gtk_dialog_yes_no',
                'defaults': ['openpaperwork_gtk.dialogs.yes_no'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GLIB_AVAILABLE:
            return

        action = Gio.SimpleAction.new(ACTION_NAME, None)
        action.connect("activate", self._delete)
        self.core.call_all("app_actions_add", action)

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def doc_open(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)

    def doc_close(self):
        self.active_doc = None

    def on_page_shown(self, page_idx):
        self.active_page_idx = page_idx

    def _delete(self, *args, **kwargs):
        assert self.active_doc is not None

        LOGGER.info(
            "Asking confirmation before deleting page %d of document %s",
            self.active_page_idx, self.active_doc[0]
        )
        msg = (
            _(
                "Are you sure you want to delete page"
                " {page_idx} of document {doc_id} ?"
            ).format(
                page_idx=(self.active_page_idx + 1),
                doc_id=self.active_doc[0]
            )
        )
        self.core.call_success(
            "gtk_show_dialog_yes_no", self, msg, self.active_doc
        )

    def on_dialog_yes_no_reply(
            self, parent, reply, *args, **kwargs):
        if parent is not self:
            return
        if not reply:
            return

        (active_doc,) = args
        (doc_id, doc_url) = active_doc
        page_idx = self.active_page_idx

        LOGGER.info("Will delete page %s p%d", doc_id, page_idx)

        self.core.call_all("page_delete_by_url", doc_url, page_idx)
        self.core.call_all("search_update_document_list")
        self.core.call_all("doc_reload", doc_id, doc_url)

        self.core.call_success("transaction_simple", [('upd', doc_id),])



"""
This plugin handles different working directories by hashing their path
and adding a part of this hash to the data directory.

If a working directly was not loaded for at least one month, it is
deleted, again.
"""
import datetime
import logging
import base64
import hashlib
import os

import openpaperwork_core


LOGGER = logging.getLogger(__name__)

WORK_DIR_NAME = "workdir_data"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return ["data_dir_handler"]

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'paths',
                'defaults': ['openpaperwork_core.paths.xdg'],
            },
            {

                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
        ]

    def init(self, core):
        super().init(core)
        self._delete_old_directories()

    def on_storage_changed(self):
        LOGGER.info(
            "Work directory has changed --> data directory has to change too"
        )
        self.core.call_all("on_data_dir_changed")

    @staticmethod
    def _hash_dir(url):
        dir_hash = hashlib.sha256(url.encode()).digest()[:6]
        encoded_hash = base64.urlsafe_b64encode(dir_hash).decode()[:8]
        return encoded_hash

    def get_dir_mtime(self, dir_url):
        mtime = 0
        file_urls = self.core.call_success(
            "fs_recurse", dir_url, dir_included=True
        )
        for file_url in file_urls:
            file_mtime = self.core.call_success("fs_get_mtime", file_url)
            if file_mtime > mtime:
                mtime = file_mtime
        return mtime

    def _delete_old_directories(self, days_to_data_dir_deletion=31):
        data_dir = self.core.call_success("paths_get_data_dir")
        work_data_dir = self.core.call_success(
            "fs_join", data_dir, WORK_DIR_NAME)
        folder_content = self.core.call_success(
            "fs_listdir", work_data_dir)
        now = datetime.datetime.now()
        for file in folder_content:
            if self.core.call_success("fs_isdir", file):
                mtime = self.get_dir_mtime(file)
                modified = datetime.datetime.fromtimestamp(mtime)
                time_diff = now - modified
                if time_diff.days >= days_to_data_dir_deletion:
                    LOGGER.info(
                        "Removing directory %s as it is older than %i days."
                        % (file, days_to_data_dir_deletion))
                    self.core.call_success("fs_rm_rf", file, trash=False)

    def data_dir_handler_get_individual_data_dir(self):
        work_dir = self.core.call_success("storage_get_id")
        data_dir = self.core.call_success("paths_get_data_dir")
        encoded_hash = Plugin._hash_dir(work_dir)
        workdir_data_folder = self.core.call_success(
            "fs_join", data_dir, WORK_DIR_NAME)
        individual_data_dir = self.core.call_success(
            "fs_join", workdir_data_folder,
            "%s_%s" % (os.path.basename(work_dir), encoded_hash)
        )
        self.core.call_success("fs_mkdir_p", individual_data_dir)
        return individual_data_dir



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import collections

import openpaperwork_core

from .. import _


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.changes = collections.defaultdict(
            # we cannot use sets here because sets are not JSON-serializable
            lambda: collections.defaultdict(list)
        )

    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                "interface": "syncable",
                "defaults": [
                    "paperwork_backend.guesswork.label.sklearn",
                    "paperwork_backend.guesswork.ocr.pyocr",
                    "paperwork_backend.index.whoosh",
                    "paperwork_backend.model.labels",
                ],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        parser.add_parser('sync', help=_(
            "Synchronize the index(es) with the content of the work directory"
        ))

    def on_sync(self, name, status, key):
        self.changes[name][status].append(key)

    def cmd_run(self, console, args):
        if args.command != 'sync':
            return None

        console.print(_("Synchronizing with work directory ..."))

        self.changes = collections.defaultdict(
            # we cannot use sets here because sets are not JSON-serializable
            lambda: collections.defaultdict(list)
        )

        self.core.call_all("transaction_sync_all")
        self.core.call_all("mainloop_quit_graceful")
        self.core.call_one("mainloop")
        console.print(_("All done !"))

        # ensure order of documents to make testing easier and ensure
        # behaviour consistency
        for actions in self.changes.values():
            for docs in actions.values():
                docs.sort()
        return dict(self.changes)



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps

from ... import _


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "page_print"


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -10

    def get_interfaces(self):
        return [
            'chkdeps',
            'menu',
            'menu_page',
            'menu_page_print',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'action_page_print',
                'defaults': ['paperwork_gtk.actions.page.print'],
            },
            {
                'interface': 'page_actions',
                'defaults': [
                    'paperwork_gtk.mainwindow.docview.pageinfo.actions'
                ],
            },
        ]

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def on_page_menu_ready(self):
        item = Gio.MenuItem.new(_("Print page"), "win." + ACTION_NAME)
        self.core.call_all("page_menu_append_item", item)



import logging
import re
import zipfile

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core

from .. import (_, deps)


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -100

    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.file_urls = []
        self.apply_handler_id = None

    def get_interfaces(self):
        return [
            'bug_report_method',
            'chkdeps',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(deps.GTK)

    def bug_report_complete(self, *args):
        url = "https://gitlab.gnome.org/World/OpenPaperwork/paperwork/issues"
        self.core.call_all(
            "bug_report_add_method",
            _("ZIP file"),
            re.sub(
                '[ \t]+', ' ',
                (
                    _(
                        "Build a ZIP file containing all the attachments.\n"
                        " If you want, you can then submit a bug report"
                        ' manually on <a href="%s">Paperwork\'s bug'
                        " tracker</a> and attach this ZIP file to the ticket."
                    ) % url
                ).strip()
            ),
            self._enable_zip, self._disable_zip,
            *args
        )

    def _enable_zip(self, bug_report_widget_tree):
        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "openpaperwork_gtk.bug_report", "bug_report_zip.glade"
        )

        assistant = bug_report_widget_tree.get_object("bug_report_dialog")
        page = self.widget_tree.get_object("bug_report_zip_page")
        assistant.append_page(page)
        assistant.set_page_complete(page, False)
        assistant.set_page_type(page, Gtk.AssistantPageType.CONFIRM)
        self.apply_handler_id = assistant.connect("apply", self._make_zip)

        filechooser = self.widget_tree.get_object("bug_report_zip_filechooser")
        filechooser.connect(
            "selection-changed",
            lambda filechooser: assistant.set_page_complete(
                page, filechooser.get_uri() is not None
            )
        )

    def _disable_zip(self, bug_report_widget_tree):
        if self.widget_tree is None:
            return
        assistant = bug_report_widget_tree.get_object("bug_report_dialog")
        page = self.widget_tree.get_object("bug_report_zip_page")
        assistant.remove(page)
        assistant.disconnect(self.apply_handler_id)
        self.widget_tree = None

    def bug_report_set_file_urls_to_send(self, file_urls):
        self.file_urls = file_urls

    def _make_zip(self, assistant):
        in_file_urls = self.file_urls
        out_file_url = self.widget_tree.get_object(
            "bug_report_zip_filechooser"
        ).get_uri()
        out_file_url = out_file_url.strip()
        if not out_file_url.lower().endswith(".zip"):
            out_file_url += ".zip"
        LOGGER.info("Adding attachments to %s", out_file_url)

        out_file_path = self.core.call_success("fs_unsafe", out_file_url)
        with zipfile.ZipFile(out_file_path, 'w') as out_zip:
            for in_file_url in in_file_urls:
                file_name = self.core.call_success("fs_basename", in_file_url)
                with self.core.call_success(
                        "fs_open", in_file_url, "rb") as in_fd:
                    with out_zip.open(file_name, 'w') as out_fd:
                        out_fd.write(in_fd.read())

        LOGGER.info("%s written", out_file_url)



import datetime
import logging
import os
import os.path
import sys
import tempfile

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


g_tmp_file = None


def _get_tmp_file():
    global g_tmp_file

    if g_tmp_file is not None:
        return g_tmp_file

    date = datetime.datetime.now()
    date = date.strftime("%Y%m%d_%H%M_%S")
    t = tempfile.NamedTemporaryFile(
        mode='w',
        suffix=".txt",
        prefix="openpaperwork_{}_".format(date),
        encoding='utf-8'
    )
    sys.stderr.write("Temporary file = {}\n".format(t.name))
    g_tmp_file = t
    return t


class _LogHandler(logging.Handler):
    def __init__(self):
        super().__init__()
        self.log_level = logging.DEBUG
        self.formatter = logging.Formatter(Plugin.DEFAULT_LOG_FORMAT)
        self.out_fds = set()
        self.out_fds.add(sys.stderr)

    def emit(self, record):
        if record.levelno < self.log_level:
            return
        line = self.formatter.format(record)
        for fd in self.out_fds:
            fd.write(line + "\n")


class Plugin(PluginBase):
    CONFIG_FILE_SEPARATOR = ","

    DEFAULT_LOG_FILES = 'stderr'
    DEFAULT_LOG_FORMAT = '[%(levelname)-6s] [%(name)-30s] %(message)s'

    LOG_LEVELS = {
        'none': logging.CRITICAL,
        'critical': logging.CRITICAL,
        'error': logging.ERROR,
        'warn': logging.WARN,
        'warning': logging.WARNING,
        'info': logging.INFO,
        'debug': logging.DEBUG,
    }
    SPECIAL_FILES = {
        'stdout': lambda: sys.stdout,
        'stderr': lambda: sys.stderr,
        'temp': _get_tmp_file,
    }

    def __init__(self):
        super().__init__()
        self.log_file_paths = set()
        self.log_handler = None

    def get_interfaces(self):
        return [
            'logs',
            'uncaught_exception_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'uncaught_exception',
                'defaults': ['openpaperwork_core.uncaught_exception'],
            },
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
        ]

    def init_logs(self, app_name, default_log_level):
        self._fix_std('stdout')
        self._fix_std('stderr')

        section_name = "logging:" + app_name

        s = self.core.call_success(
            "config_build_simple", section_name, "level",
            lambda: default_log_level
        )
        self.core.call_all("config_register", "log_level", s)
        self.core.call_all(
            'config_add_observer', "log_level", self._reload_config
        )

        s = self.core.call_success(
            "config_build_simple", section_name, "files",
            lambda: self.DEFAULT_LOG_FILES
        )
        self.core.call_all("config_register", "log_files", s)
        self.core.call_all(
            'config_add_observer', "log_files", self._reload_config
        )

        s = self.core.call_success(
            "config_build_simple", section_name, "format",
            lambda: self.DEFAULT_LOG_FORMAT
        )
        self.core.call_all("config_register", "log_format", s)
        self.core.call_all(
            'config_add_observer', "log_format", self._reload_config
        )

        self.log_handler = _LogHandler()
        logging.getLogger().addHandler(self.log_handler)

        self._reload_config()

    def _fix_std(self, std_stream_name: str):
        if getattr(sys, std_stream_name, None) is not None:
            return
        # When frozen, sys.stderr / sys.stdout == None which is a problem
        if os.path.exists("/dev/null"):
            std_stream = open("/dev/null", "w")  # Linux / MacOS
        else:
            std_stream = open("NULL", "w")  # Windows
        setattr(sys, std_stream_name, std_stream)

    def _disable_logging(self):
        for fd in self.log_handler.out_fds:
            if fd != sys.stdout and fd != sys.stderr and fd != g_tmp_file:
                fd.close()
        self.log_handler.out_fds = set()

    def _enable_logging(self):
        self.log_handler.out_fds = set()
        first = None
        for file_path in self.log_file_paths:
            if file_path.lower() not in self.SPECIAL_FILES:
                fd = open(file_path, 'a')
                self.log_handler.out_fds.add(fd)
            else:
                fd = self.SPECIAL_FILES[file_path.lower()]()
                self.log_handler.out_fds.add(fd)
            if first is None and fd is not None:
                first = fd

    def _reload_config(self, *args, **kwargs):
        self._disable_logging()
        LOGGER.info("Reloading logging configuration")
        try:
            log_level = self.core.call_success('config_get', "log_level")
            self.set_log_level(log_level)
            self.log_file_paths = self.core.call_success(
                'config_get', "log_files",
            ).split(self.CONFIG_FILE_SEPARATOR)
            self.log_handler.formatter = logging.Formatter(
                self.core.call_success('config_get', "log_format")
            )
        finally:
            self._enable_logging()

    def set_log_level(self, log_level):
        lvl = self.LOG_LEVELS[log_level]
        self.log_handler.log_level = lvl
        if lvl > logging.INFO:
            # Never disable info level ; it may be used by other plugins
            lvl = logging.INFO
        logging.getLogger().setLevel(lvl)

    def on_uncaught_exception(self, exc_info):
        LOGGER.error("=== UNCAUGHT EXCEPTION ===", exc_info=exc_info)
        LOGGER.error("==========================")
